import re
import numpy as np
from typing import Dict, Any, Optional, Tuple

import textarena as ta
# from textarena.envs.IteratedStagHunt.renderer import create_board_str


class IteratedStagHuntEnv(ta.Env):
    def __init__(self, num_rounds: int=5, conversation_rounds: int=3, mutual_stag_reward: int=10, single_hare_reward: int=8, single_stag_reward: int=1, mutual_hare_reward: int=5, randomize_payoff: bool=False):
        self.num_rounds = num_rounds
        self.conversation_rounds = conversation_rounds
        
        # payoffs
        self.mutual_stag_reward = mutual_stag_reward
        self.single_hare_reward = single_hare_reward
        self.single_stag_reward = single_stag_reward
        self.mutual_hare_reward = mutual_hare_reward
        self.randomize_payoff = randomize_payoff

        # Action patterns
        self.stag_pattern = re.compile(r"\[Stag\]", re.IGNORECASE)
        # self.hare_pattern = re.compile(r"\[Hare\]", re.IGNORECASE)

        self.curr_payoff: Dict[str, int] = {}


    def reset(self, num_players: int, seed: Optional[int] = None):
        self.state = ta.TwoPlayerState(num_players=num_players, seed=seed)
        game_state = {
            "round": 1, "num_rounds": self.num_rounds, "phase": "conversation", "conversation_round": 0, "total_conversation_rounds": self.conversation_rounds, 
            "decisions": {0: None, 1: None}, "total_payoff": {0: 0, 1: 0}
        }
        self.state.reset(game_state=game_state, player_prompt_function=self._prompt)
        self._create_round_payoff_matrix() 
    
    def _prompt(self, player_id: int, game_state: Dict[str, Any]) -> str:
        """Generate the initial prompt for a player."""
        return (
            f"You are Player {player_id} in an {game_state['num_rounds']} round game of Iterated Stag Hunt.\n\n"
            f"Game Structure:\n"
            f"- The game consists of {self.num_rounds} decision rounds\n"
            f"- Before each decision, you have {game_state['total_conversation_rounds']} turns to communicate\n"
            f"- After communication, both players simultaneously choose to hunt a Stag or Hare\n\n"
            f"Rewards:\n"
            f"- The rewards associated with hunting stags and hares may differ between rounds\n"
            f"- The rewards are presented at the start of each round\n\n"
            f"How to Play:\n"
            f"- During communication: Simply type your message\n"
            f"- During decision phase: Use '[Stag]' or '[Hare]'\n"
            f"You can include additional text before or after these tokens.\n"
        )
    
    def _create_round_payoff_matrix(self) -> None:
        if not self.randomize_payoff:
            self.mutual_stag_payoff = self.mutual_stag_reward
            self.single_stag_payoff = self.single_stag_reward
            self.single_hare_payoff = self.single_hare_reward
            self.mutual_hare_payoff = self.mutual_hare_reward
        else:
            self.single_stag_payoff = self.single_stag_reward
            self.mutual_hare_payoff = np.random.randint(self.single_stag_payoff+1, self.mutual_hare_reward+1)
            self.single_hare_payoff = np.random.randint(self.mutual_hare_payoff, self.single_hare_reward+1)
            self.mutual_stag_payoff = np.random.randint(self.single_hare_payoff+1, self.mutual_stag_reward+1)

        message = (
            f"Starting Round {self.state.game_state['round']} with payoff matrix:\n"
            f"- Both hunt a Stag: Both get {self.mutual_stag_payoff} points\n"
            f"- Both hunt a Hare: Both get {self.mutual_hare_payoff} points\n"
            f"- One hunts a Hare, One hunts a Stag: The hunter of the Hare gets {self.mutual_hare_payoff} points, the hunter of the Stag gets {self.single_stag_payoff} points\n"
            f"You can freely communicate with your opponent for {self.state.game_state['total_conversation_rounds']} rounds before making a decision."
        )
        self.state.add_observation(message=message, observation_type=ta.ObservationType.GAME_MESSAGE)

    def step(self, action: str) -> Tuple[bool, Dict]:
        self.state.add_observation(to_id=self.state.current_player_id, from_id=self.state.current_player_id, message=action, observation_type=ta.ObservationType.PLAYER_ACTION)

        # check what phase we are in
        match self.state.game_state["phase"]:
            case "conversation":    self._handle_conversation_phase(action=action)
            case "decision":        self._handle_decision_phase(action=action)
        return self.state.step()

    def _handle_conversation_phase(self, action: str):
        # broadcast player action to opponent
        self.state.add_observation(to_id=1-self.state.current_player_id, from_id=self.state.current_player_id, message=action.strip(), observation_type=ta.ObservationType.PLAYER_ACTION)

        # only increment conversation round on player 1
        if self.state.current_player_id == 1:
            self.state.game_state["conversation_round"] += 1

            # check if we have completed all conversation rounds
            if self.state.game_state["conversation_round"] >= self.state.game_state["total_conversation_rounds"]:
                self.state.game_state["phase"] = "decision" # rotate phase
                self.state.add_observation(message=f"The decision phase for round {self.state.game_state['round']} has started. Please reply with either '[stag]' or '[hare]'.", observation_type=ta.ObservationType.GAME_BOARD) # send decision prompt to everybody

    def _handle_decision_phase(self, action: str):
        # extraction decision
        self.state.game_state["decisions"][self.state.current_player_id] = "hare" if self.stag_pattern.search(action) else "stag" # stag if no action
        
        # check if we have both decisions
        if all(decision is not None for decision in self.state.game_state["decisions"].values()):
            # determine pay-off & send message
            round_payoffs = [None, None]
            if self.state.game_state["decisions"][0] == self.state.game_state["decisions"][1]: # matching decision 
                round_payoffs[0] = round_payoffs[1] = (self.mutual_stag_payoff if self.state.game_state["decisions"][0]=="stag" else self.mutual_hare_payoff)
            else: # differing decisions
                round_payoffs[0] = self.single_stag_payoff if self.state.game_state["decisions"][0]=="stag" else self.single_hare_payoff
                round_payoffs[1] = self.single_stag_payoff if self.state.game_state["decisions"][1]=="stag" else self.single_hare_payoff
            
            message=f"Round {self.state.game_state['round']} complete. Results:"
            for pid in [0,1]:
                self.state.game_state['total_payoff'][pid] += round_payoffs[pid]
                message+= f"\n\tPlayer {pid} picked '{self.state.game_state["decisions"][pid]}' (payoff: {round_payoffs[pid]}; total: {self.state.game_state['total_payoff'][pid]})"
            self.state.add_observation(message=message, observation_type=ta.ObservationType.GAME_MESSAGE)


            # start next round
            self.state.game_state["round"] += 1
            self.state.game_state["phase"] = "conversation"; self.state.game_state["conversation_round"] = 0; self.state.game_state["decisions"] = {0: None, 1: None} # reset
            self._create_round_payoff_matrix()
            
            # check if round limit reached
            if self.state.game_state["round"] >= self.state.game_state["num_rounds"]:
                # determine winner and return
                self._determine_winner()

    def _determine_winner(self):
        if self.state.game_state["total_payoff"][0] > self.state.game_state["total_payoff"][1]:     winner_id = 0
        elif self.state.game_state["total_payoff"][1] > self.state.game_state["total_payoff"][0]:   winner_id = 1
        else:                                                                                       winner_id = None
        reason = f"Final scores: Player 0: {self.state.game_state['total_payoff'][0]}, Player 1: {self.state.game_state['total_payoff'][1]}\n"+ (f"It's a Tie!" if winner_id == None else f"Player {winner_id} won!")
        self.state.set_winner(player_id=winner_id, reason=reason)




import re, random, importlib
from typing import Any, Union, List, Callable, Dict, Tuple, Optional
from dataclasses import dataclass, field

import textarena as ta 


# Global environment registry
ENV_REGISTRY: Dict[str, Callable] = {}

@dataclass
class EnvSpec:
    """A specification for creating environments."""
    id: str
    entry_point: Callable
    default_wrappers: Optional[List[ta.Wrapper]]
    kwargs: Dict[str, Any] = field(default_factory=dict)
    
    def make(self, **kwargs) -> Any:
        """Create an environment instance."""
        all_kwargs = {**self.kwargs, **kwargs}
        return self.entry_point(**all_kwargs)
    
def register(id: str, entry_point: Callable, default_wrappers: Optional[List[ta.Wrapper]]=None, **kwargs: Any):
    """Register an environment with a given ID."""
    if id in ENV_REGISTRY:
        raise ValueError(f"Environment {id} already registered.")
    ENV_REGISTRY[id] = EnvSpec(id=id, entry_point=entry_point, default_wrappers=default_wrappers, kwargs=kwargs)

def register_with_versions(id: str, entry_point: Callable, wrappers: Optional[Dict[str, List[ta.Wrapper]]]=None, **kwargs: Any):
    """Register an environment with a given ID."""
    if id in ENV_REGISTRY: raise ValueError(f"Environment {id} already registered.")

    # first register default version
    ENV_REGISTRY[id] = EnvSpec(id=id, entry_point=entry_point, default_wrappers=wrappers.get("default"), kwargs=kwargs)
    for wrapper_version_key in list(wrappers.keys())+["-raw"]:
        if wrapper_version_key=="default": continue
        ENV_REGISTRY[f"{id}{wrapper_version_key}"] = EnvSpec(id=f"{id}{wrapper_version_key}", entry_point=entry_point, default_wrappers=wrappers.get(wrapper_version_key), kwargs=kwargs)

def pprint_registry_detailed():
    """Pretty print the registry with additional details like kwargs."""
    if not ENV_REGISTRY:
        print("No environments registered.")
    else:
        print("Detailed Registered Environments:")
        for env_id, env_spec in ENV_REGISTRY.items():
            print(f"  - {env_id}:")
            print(f"      Entry Point: {env_spec.entry_point}")
            print(f"      Kwargs:      {env_spec.kwargs}")
            print(f"      Wrappers:    {env_spec.default_wrappers}")

def check_env_exists(env_id: str):
    """Check if an environment exists in the registry."""
    if env_id not in ENV_REGISTRY:
        raise ValueError(f"Environment {env_id} is not registered.")
    else:
        print(f"Environment {env_id} is registered.")

def make(env_id: Union[str, List[str]], **kwargs) -> Any:
    """Create an environment instance using the registered ID."""
    # If env_id is a list, randomly select one environment ID
    if isinstance(env_id, list):
        if not env_id:
            raise ValueError("Empty list of environment IDs provided.")
        env_id = random.choice(env_id)
    
    # Continue with the existing implementation
    if env_id not in ENV_REGISTRY:
        raise ValueError(f"Environment {env_id} not found in registry.")
    
    env_spec = ENV_REGISTRY[env_id]
    
    # Resolve the entry point if it's a string
    if isinstance(env_spec.entry_point, str):
        module_path, class_name = env_spec.entry_point.split(":")
        try:
            module = importlib.import_module(module_path)
            env_class = getattr(module, class_name)
        except (ModuleNotFoundError, AttributeError) as e:
            raise ImportError(f"Could not import {module_path}.{class_name}. Error: {e}")
    else:
        env_class = env_spec.entry_point
    
    env = env_class(**{**env_spec.kwargs, **kwargs})

    # Dynamically attach the env_id
    env.env_id = env_id
    env.entry_point = env_spec.entry_point

    # wrap the environment
    if env_spec.default_wrappers is not None and len(env_spec.default_wrappers) > 0:
        for wrapper in env_spec.default_wrappers:
            env = wrapper(env)

    return env



import textarena as ta 
from textarena.core import ObservationWrapper, Env, Observations, Info, ObservationType
from typing import Dict, Optional, Tuple, List

__all__ = ["LLMObservationWrapper", "DiplomacyObservationWrapper", "FirstLastObservationWrapper", "FirstLastObservationWrapper"]


class LLMObservationWrapper(ObservationWrapper):
    """
    A wrapper for converting environment observations into formatted strings suitable
    for large language models (LLMs). It ensures that duplicate observations are not
    added to the full observations.
    """

    def __init__(self, env: Env):
        """
        Initializes the LLMObservationWrapper.

        Args:
            env (Env): The environment to wrap.
        """
        super().__init__(env)
        self.full_observations: Dict[int, List[Tuple[int, str]]] = {}

    def _convert_obs_to_str(self, player_id: int) -> Observations:
        """
        Converts the full observations into formatted strings for each recipient.

        Returns:
            Observations: A dictionary mapping recipient IDs to their formatted observation strings.
        """
        str_observation = ""
        
        if player_id in self.full_observations:
            for sender_id, message, _ in self.full_observations[player_id]:
                if sender_id == ta.GAME_ID:
                    sender_name = "GAME"
                else:
                    sender_name = self.env.state.role_mapping.get(sender_id, f"Player {sender_id}")
                str_observation += f"\n[{sender_name}] {message}"

        return str_observation

    def observation(self, player_id: int, observation: Optional[ta.Observations]):
        if observation is None:
            return self._convert_obs_to_str(player_id=player_id)

        # Extend the full observations with the current observations without duplicates
        if player_id not in self.full_observations:
            self.full_observations[player_id] = []

        # Append new observations in sequence
        self.full_observations[player_id].extend(observation)

        return self._convert_obs_to_str(player_id=player_id)


    
class DiplomacyObservationWrapper(LLMObservationWrapper):
    def __init__(self, env: ta.Env):
        super().__init__(env)

    def _get_history_conversation(self, player_id: int) -> str:
        """ Get the history conversation for the given player. """
        history = []
        for sender_id, message, _ in self.full_observations[player_id][1:]:
            if sender_id == ta.GAME_ID:
                sender_name = "GAME"
            else:
                sender_name = self.env.state.role_mapping.get(sender_id, f"Player {sender_id}")
            history.append(f"[{sender_name}] {message}")
        return "\n".join(history)

    def observation(self, player_id: int, observation: Optional[ta.Observations]):
        if observation is None:
            return self.env.get_prompt(player_id, self._get_history_conversation(player_id))

        if player_id not in self.full_observations:
            self.full_observations[player_id] = []

        self.full_observations[player_id].extend(observation)

        return self.env.get_prompt(player_id, self._get_history_conversation(player_id))


class FirstLastObservationWrapper(ObservationWrapper):
    def __init__(self, env: Env):
        super().__init__(env)
        self.full_observations: Dict[int, List[Tuple[int, str]]] = {}

    def _convert_obs_to_str(self, player_id: int) -> Observations:
        return_str = self.full_observations[player_id][0][1]
        if len(self.full_observations[player_id]) > 1:
            return_str += "\n\n" + self.full_observations[player_id][-1][1]

        return return_str + "\n\n" + "Next Action:"

    def observation(self, player_id: int, observation: Optional[ta.Observations]):
        if observation is None:
            return self._convert_obs_to_str(player_id=player_id)

        # Extend the full observations with the current observations without duplicates
        if player_id not in self.full_observations:
            self.full_observations[player_id] = []

        # Append new observations in sequence
        self.full_observations[player_id].extend(observation)

        return self._convert_obs_to_str(player_id=player_id)
    

class GameBoardObservationWrapper(ObservationWrapper):
    """ show the initial prompt and the most recent game board """
    def __init__(self, env: Env):
        super().__init__(env)
        self.full_observations: Dict[int, List[Tuple[int, str]]] = {}

    def _convert_obs_to_str(self, player_id: int) -> Observations:
        prompt = [obs[1] for obs in self.full_observations[player_id] if obs[2] == ObservationType.PROMPT][0]
        last_board_state = [obs[1] for obs in self.full_observations[player_id] if obs[2] == ObservationType.GAME_BOARD][-1]
        assert prompt and last_board_state, f"You are using the GameBoardObservationWrapper, but either the ObservationType.PROMPT or ObservationType.GAME_BOARD is missing"
        return prompt + "\n\n" + last_board_state # + "\n\n" + "Next Action:"

    def observation(self, player_id: int, observation: Optional[ta.Observations]):
        if observation is None: return self._convert_obs_to_str(player_id=player_id)
        # Extend the full observations with the current observations without duplicates
        if player_id not in self.full_observations: self.full_observations[player_id] = []
        self.full_observations[player_id].extend(observation) # Append new observations in sequence
        return self._convert_obs_to_str(player_id=player_id)

    


class GameMessagesObservationWrapper(ObservationWrapper):
    """ show the initial prompt and the messages sent by the game """
    def __init__(self, env: Env):
        super().__init__(env)
        self.full_observations: Dict[int, List[Tuple[int, str]]] = {}

    def _convert_obs_to_str(self, player_id: int) -> Observations:
        str_observation = ""
        if player_id in self.full_observations:
            for _, message, observation_type in self.full_observations[player_id]:
                if observation_type not in [ObservationType.PLAYER_ACTION, ObservationType.GAME_ADMIN]: 
                    str_observation += f"\n {message}"
        return str_observation

    def observation(self, player_id: int, observation: Optional[ta.Observations]):
        if observation is None: return self._convert_obs_to_str(player_id=player_id)
        # Extend the full observations with the current observations without duplicates
        if player_id not in self.full_observations: self.full_observations[player_id] = []
        self.full_observations[player_id].extend(observation) # Append new observations in sequence
        return self._convert_obs_to_str(player_id=player_id)

class GameMessagesAndCurrentBoardObservationWrapper(ObservationWrapper):
    """Show the initial prompt, game messages (excluding player actions), and current game board."""
    def __init__(self, env: Env):
        super().__init__(env)
        self.full_observations: Dict[int, List[Tuple[int, str, ObservationType]]] = {}

    def _convert_obs_to_str(self, player_id: int) -> Observations:
        str_observation = ""
        prompt = None
        board_state = None
        for _, message, obs_type in self.full_observations.get(player_id, []):
            if obs_type == ObservationType.PROMPT:
                prompt = message
            elif obs_type == ObservationType.GAME_BOARD:
                board_state = message
            elif obs_type not in [ObservationType.PLAYER_ACTION, ObservationType.GAME_ADMIN]:
                str_observation += f"\n{message}"

        if prompt is None or board_state is None:
            raise ValueError("Missing required observation types: PROMPT or GAME_BOARD")

        return f"{prompt}\n\n{str_observation.strip()}\n\n{board_state}" #\n\nNext Action:"

    def observation(self, player_id: int, observation: Optional[ta.Observations]):
        if observation is None:
            return self._convert_obs_to_str(player_id=player_id)

        if player_id not in self.full_observations:
            self.full_observations[player_id] = []

        self.full_observations[player_id].extend(observation)
        return self._convert_obs_to_str(player_id=player_id)

# class GameMessagesAndCurrentBoardWithInvalidMovesObservationWrapper(ObservationWrapper):
#     """Show the initial prompt, game messages (excluding player actions), and current game board."""
#     def __init__(self, env: Env):
#         super().__init__(env)
#         self.full_observations: Dict[int, List[Tuple[int, str, ObservationType]]] = {}

#     def _convert_obs_to_str(self, player_id: int) -> Observations:
#         prompt, board_state, story = None, None, []
#         for _, message, obs_type in self.full_observations.get(player_id, []):
#             if obs_type == ObservationType.PROMPT:
#                 prompt = message
#             elif obs_type == ObservationType.GAME_BOARD:
#                 board_state = message
#             elif obs_type not in (ObservationType.PLAYER_ACTION, ObservationType.GAME_ADMIN):
#                 story.append(message)

#         if prompt is None or board_state is None:
#             raise ValueError("Missing required observation types: PROMPT or GAME_BOARD")

#         obs_list = self.full_observations.get(player_id, [])

#         last_pa_idx = None
#         for idx in range(len(obs_list) - 1, -1, -1):
#             if obs_list[idx][2] == ObservationType.PLAYER_ACTION:
#                 last_pa_idx = idx
#                 break

#         prev_action_msg, admin_msg = "", None
#         if last_pa_idx is not None: 
#             prev_action_msg = obs_list[last_pa_idx][1]
#             for _, msg, typ in obs_list[last_pa_idx + 1 :]:
#                 if typ == ObservationType.GAME_ADMIN:
#                     admin_msg = msg
#         return_str = f"{prompt}\n\n{'\n'.join(story).strip()}\n\n{board_state}"
#         if admin_msg is not None: return_str += f"\n\nYour previous action: '{prev_action_msg[-250:]}'\n\n{admin_msg}"

#         return return_str

    def observation(self, player_id: int, observation: Optional[ta.Observations]):
        if observation is None:
            return self._convert_obs_to_str(player_id=player_id)

        if player_id not in self.full_observations:
            self.full_observations[player_id] = []

        self.full_observations[player_id].extend(observation)
        return self._convert_obs_to_str(player_id=player_id)

class SingleTurnObservationWrapper(ObservationWrapper):
    def __init__(self, env: Env):
        super().__init__(env)

    def observation(self, player_id: int, observation: Optional[ta.Observations]):
        return observation[0][1]



import re
from typing import Optional, Dict, Tuple, List, Any

import textarena as ta
from textarena.envs.Othello.renderer import create_board_str


DIRS = [(-1,-1),(-1,0),(-1,1),(0,-1),(0,1),(1,-1),(1,0),(1,1)]
EMPTY, BLACK, WHITE = "", "B", "W"

class OthelloEnv(ta.Env):
    def __init__(self, board_size: int = 8, show_valid: bool = True):
        if board_size % 2 or board_size < 4: raise ValueError("board_size must be an even integer ≥ 4")
        super().__init__()
        self.N = board_size
        self.show_valid = show_valid

    def _in_bounds(self, r: int, c: int) -> bool:
        return 0 <= r < self.N and 0 <= c < self.N

    def _counts(self) -> Tuple[int, int]:
        b = sum(row.count(BLACK) for row in self.board)
        w = sum(row.count(WHITE) for row in self.board)
        return b, w

    def get_board_str(self) -> str:
        return create_board_str(self.board)

    def reset(self, num_players: int, seed: Optional[int] = None):
        self.state = ta.TwoPlayerState(num_players=num_players, seed=seed)
        self.board = [[EMPTY for _ in range(self.N)] for _ in range(self.N)] # empty board

        # initial four stones in the middle
        m1, m2 = self.N // 2 - 1, self.N // 2
        self.board[m1][m1] = self.board[m2][m2] = WHITE
        self.board[m1][m2] = self.board[m2][m1] = BLACK

        b_count, w_count = self._counts()
        valid_moves = self._valid_moves(BLACK)

        game_state={"board": self.board, "rendered_board": self._render_board(), "black_count": b_count, "white_count": w_count, "valid_moves": valid_moves}
        self.state.reset(game_state=game_state, player_prompt_function=self._prompt, role_mapping={0: "Black", 1: "White"})

        obs = f"Game Board:\n{self.state.game_state['rendered_board']}"
        if self.show_valid: obs += "\nValid moves: " + ", ".join([f"'{vm}'" for vm in valid_moves]) if valid_moves else "\nNo valid moves - you may have to skip."
        obs += f"\nCurren Scores - Black: {self.state.game_state['black_count']}, White: {self.state.game_state['white_count']}\n"
        self.state.add_observation(message=obs, observation_type=ta.ObservationType.GAME_BOARD)

    def _prompt(self, player_id: int, game_state: Dict[str, Any]) -> str:
        piece, colour = (BLACK, "Black") if player_id == 0 else (WHITE, "White")
        return (
            f"You are Player {player_id} playing {colour} ({piece}) in a game of Othello.\nYour goal is to have more pieces of your color on the board by the end of the game.\n"
            f"On your turn, place a piece such that it flanks one or more of your opponent's pieces-in any direction (horizontal, vertical, or diagonal)-between your new piece and another of your existing pieces. "
            f"All flanked opponent pieces will be flipped to your color.\nProvide your move in the format '[row, col]'."
        )

    def step(self, action: str) -> Tuple[bool, ta.Info]:
        pid = self.state.current_player_id
        piece = BLACK if pid == 0 else WHITE
        opp = BLACK if piece == WHITE else WHITE
        self.state.add_observation(from_id=pid, message=action, observation_type=ta.ObservationType.PLAYER_ACTION)

        valid = self._valid_moves(piece)
        if not valid:
            self._handle_skip(pid, piece, opp)
            obs = f"Player {pid} had to skip their turn" 
        else:
            match = re.compile(r"\[\s*(\d+)\s*,?\s*(\d+)\s*\]").search(action)
            if match is None:
                self.state.set_invalid_move(reason="Move must be of the form [row, col].")
                return self.state.step(rotate_player=False)

            r, c = map(int, match.groups())
            if [r, c] not in valid:
                self.state.set_invalid_move(reason=f"Illegal move. Valid moves: {valid}")
                return self.state.step(rotate_player=False)

            flipped = self._place_and_flip(r, c, piece)
            obs = f"Player {pid} ({piece}) played [{r}, {c}] flipping {flipped} piece(s)"
            
            
        self.state.add_observation(message=obs, observation_type=ta.ObservationType.GAME_ACTION_DESCRIPTION)

        next_valid = self._valid_moves(opp)
        self.state.game_state["valid_moves"] = next_valid
        self._push_gamestate()

        obs = f"Game Board:\n{self.state.game_state['rendered_board']}"
        if self.show_valid:
            obs += "\nValid moves: " + ", ".join([f"'{vm}'" for vm in next_valid]) if next_valid else "\nNo valid moves – you may have to skip."
        obs += f"\nScores - Black: {self.state.game_state['black_count']}, White: {self.state.game_state['white_count']}\n"
        self.state.add_observation(message=obs, observation_type=ta.ObservationType.GAME_BOARD)

        if self._game_over(): self._declare_winner()
        else: self.state.game_state["valid_moves"] = next_valid
        return self.state.step()

    def _handle_skip(self, pid, piece, opp):
        self.state.add_observation(f"Player {pid} ({piece}) has no valid moves and must skip.", observation_type=ta.ObservationType.GAME_MESSAGE)
        if not self._valid_moves(opp):
            self._declare_winner()
        else:
            self.state.game_state["valid_moves"] = self._valid_moves(opp)

    def _valid_moves(self, piece) -> List[List[int]]:
        return [[r, c] for r in range(self.N) for c in range(self.N) if self.board[r][c] == EMPTY and self._would_flip(r, c, piece)]

    def _would_flip(self, r, c, piece) -> bool:
        opp = BLACK if piece == WHITE else WHITE
        for dr, dc in DIRS:
            rr, cc = r + dr, c + dc
            if not (self._in_bounds(rr, cc) and self.board[rr][cc] == opp): continue
            while self._in_bounds(rr, cc) and self.board[rr][cc] == opp:
                rr += dr; cc += dc
            if self._in_bounds(rr, cc) and self.board[rr][cc] == piece: return True
        return False

    def _place_and_flip(self, r, c, piece) -> int:
        opp = BLACK if piece == WHITE else WHITE
        self.board[r][c] = piece
        flipped = 0
        for dr, dc in DIRS:
            rr, cc = r + dr, c + dc
            line: List[Tuple[int, int]] = []
            while self._in_bounds(rr, cc) and self.board[rr][cc] == opp:
                line.append((rr, cc))
                rr += dr; cc += dc
            if self._in_bounds(rr, cc) and self.board[rr][cc] == piece:
                for fr, fc in line:
                    self.board[fr][fc] = piece
                flipped += len(line)
        return flipped

    def _game_over(self) -> bool:
        return all(cell != EMPTY for row in self.board for cell in row) or (not self._valid_moves(BLACK) and not self._valid_moves(WHITE))

    def _render_board(self) -> str:
        header = "  " + " ".join(map(str, range(self.N)))
        rows = [header]
        for i, row in enumerate(self.board):
            cells = [cell if cell else "." for cell in row]
            rows.append(f"{i}|" + "|".join(cells) + "|")
        return "\n".join(rows)

    def _push_gamestate(self):
        b, w = self._counts()
        self.state.game_state.update({"rendered_board": self._render_board(), "black_count": b, "white_count": w})

    def _declare_winner(self):
        b, w = self._counts()
        if b > w: self.state.set_winner(player_id=0, reason=f"Black wins {b}-{w}.")
        elif w > b: self.state.set_winner(player_id=1, reason=f"White wins {w}-{b}.")
        else: self.state.set_draw(reason=f"Draw {b}-{w}.")




def create_game_str(game_state: dict) -> str:
    """Create a formatted string representation of the Colonel Blotto game state"""
    
    lines = []
    
    # Header
    current_round = game_state.get('current_round', 1)
    max_rounds = 10  # Default, should be passed in but keeping simple for now
    phase = game_state.get('phase', 'allocation')
    scores = game_state.get('scores', {0: 0, 1: 0})
    
    lines.append("┌" + "─" * 50 + "┐")
    lines.append(f"│{'COLONEL BLOTTO':^50}│")
    lines.append(f"│{f'Round {current_round} - {phase.title()} Phase':^50}│")
    lines.append("├" + "─" * 50 + "┤")
    lines.append(f"│{f'Score: Alpha {scores[0]} - Beta {scores[1]}':^50}│")
    lines.append("├" + "─" * 50 + "┤")
    
    # Fields display
    fields = game_state.get('fields', [])
    if fields:
        lines.append("│" + " " * 50 + "│")
        lines.append(f"│{'BATTLEFIELD':^50}│")
        lines.append("│" + " " * 50 + "│")
        
        # Field headers
        field_header = "│ Field │ Alpha │ Beta  │ Winner    │"
        lines.append("├" + "─" * 7 + "┼" + "─" * 7 + "┼" + "─" * 7 + "┼" + "─" * 11 + "┤")
        lines.append(field_header)
        lines.append("├" + "─" * 7 + "┼" + "─" * 7 + "┼" + "─" * 7 + "┼" + "─" * 11 + "┤")
        
        # Field data
        for field in fields:
            name = field['name']
            alpha_units = field['player_0_units']
            beta_units = field['player_1_units']
            
            if alpha_units > beta_units:
                winner = "🟢 Alpha"
            elif beta_units > alpha_units:
                winner = "🔴 Beta"
            else:
                winner = "⚪ Tie"
            
            # Handle case where no units are allocated yet
            if alpha_units == 0 and beta_units == 0 and phase == 'allocation':
                winner = "❓ TBD"
            
            field_row = f"│   {name:^3} │  {alpha_units:^3}  │  {beta_units:^3}  │ {winner:^9} │"
            lines.append(field_row)
    
    # Player states (if in allocation phase)
    if phase == 'allocation':
        lines.append("├" + "─" * 50 + "┤")
        lines.append("│" + " " * 50 + "│")
        lines.append(f"│{'ALLOCATION STATUS':^50}│")
        lines.append("│" + " " * 50 + "│")
        
        player_states = game_state.get('player_states', {})
        for player_id in [0, 1]:
            player_name = "Alpha" if player_id == 0 else "Beta"
            if player_id in player_states:
                state = player_states[player_id]
                remaining = state.get('units_remaining', 0)
                complete = state.get('allocation_complete', False)
                
                if complete:
                    status = "✅ Complete"
                    allocation = state.get('current_allocation', {})
                    alloc_str = ", ".join([f"{k}:{v}" for k, v in allocation.items()])
                    if len(alloc_str) > 35:
                        alloc_str = alloc_str[:32] + "..."
                    lines.append(f"│ {player_name}: {status} - {alloc_str:<25} │")
                else:
                    status = f"⏳ {remaining} units left"
                    lines.append(f"│ {player_name}: {status:<40} │")
    
    # Instructions (if in allocation phase and game not over)
    if phase == 'allocation':
        lines.append("├" + "─" * 50 + "┤")
        lines.append("│" + " " * 50 + "│")
        lines.append(f"│{'Enter allocation: A:4, B:2, C:2':^50}│")
        lines.append("│" + " " * 50 + "│")
    
    lines.append("└" + "─" * 50 + "┘")
    
    return "\n".join(lines)


def create_simple_game_str(game_state: dict) -> str:
    """Create a simple text representation for debugging"""
    phase = game_state.get('phase', 'allocation')
    current_round = game_state.get('current_round', 1)
    scores = game_state.get('scores', {0: 0, 1: 0})
    
    lines = [
        f"=== COLONEL BLOTTO - Round {current_round} ===",
        f"Phase: {phase.title()}",
        f"Score: Alpha {scores[0]} - Beta {scores[1]}",
        ""
    ]
    
    fields = game_state.get('fields', [])
    if fields and phase == 'results':
        lines.append("Battle Results:")
        for field in fields:
            alpha = field['player_0_units']
            beta = field['player_1_units']
            winner = "Alpha" if alpha > beta else "Beta" if beta > alpha else "Tie"
            lines.append(f"  Field {field['name']}: Alpha {alpha} vs Beta {beta} -> {winner}")
    
    return "\n".join(lines)



def create_board_str(game_state: dict) -> str:
    board = game_state["board"]
    grid_size = len(board[0])

    def cell_repr(cell):
        if cell == '~':
            return "🌊"
        elif cell == 'X':
            return "🔥"
        elif cell == 'O':
            return "💦"
        else:
            return f"{cell} "

    def render_single_board(grid, title):
        lines = []
        header = "    " + "   ".join(str(i) for i in range(grid_size))
        lines.append(f"    {title.center(grid_size * 5 - 1)}")
        lines.append("  ┌" + "────┬" * (grid_size - 1) + "────┐")
        for r in range(grid_size):
            row_cells = " │ ".join(cell_repr(grid[r][c]) for c in range(grid_size))
            lines.append(f"{chr(65 + r)} │ {row_cells} │")
            if r < grid_size - 1:
                lines.append("  ├" + "────┼" * (grid_size - 1) + "────┤")
            else:
                lines.append("  └" + "────┴" * (grid_size - 1) + "────┘")
        return lines

    board_0 = render_single_board(board[0], "Player 0")
    board_1 = render_single_board(board[1], "Player 1")

    # Merge the boards side by side
    combined = []
    for left, right in zip(board_0, board_1):
        combined.append(left + "    " + right)

    return "\n".join(combined)


import random, re
from typing import Dict, Any, Optional, Tuple

import textarena as ta
# from textarena.envs.GameOfPureStrategy.renderer import create_board_str  # TODO

class GameOfPureStrategyEnv(ta.Env):
    def __init__(self):
        super().__init__()
        self.full_hand = list(range(1, 14))
        self.action_space = re.compile(r"\[(a|k|q|j|10|[2-9])]", re.I)

    @staticmethod
    def _face_to_val(face: str) -> int:
        face = face.strip().lower()
        faces = {"a": 1, "j": 11, "q": 12, "k": 13}
        if face.isdigit(): return int(face)
        return faces.get(face)
    
    @staticmethod
    def _val_to_face(v: int) -> str: return {1: "A", 11: "J", 12: "Q", 13: "K"}.get(v, str(v))
    # def get_board_str(self): return create_board_str(self.state.game_state) # TODO

    def reset(self, num_players: int, seed: Optional[int] = None):
        self.state = ta.TwoPlayerState(num_players=num_players, seed=seed)
        game_state={
            "round": 0, "prize_deck": random.sample(self.full_hand, k=13), "carry_pot": 0, "current_prize": None, 
            "player_hands": {0: self.full_hand.copy(), 1: self.full_hand.copy()}, "pending_bids": {}, "player_scores": {0: 0, 1: 0}, "starting_player": 0,
            }
        self.state.reset(game_state=game_state, player_prompt_function=self._prompt)
        self._start_round()

    def _prompt(self, player_id: int, game_state: Dict[str, Any]) -> str:
        return (
            f"You are Player {player_id} in a match of GameOfPureStrategy.\n- You hold the 13 cards A-K; each can be used once.\n- Each round a prize card is revealed. Play exactly ONE card "
            f"by writing something that contains a bracketed token like '[Q]', '[10]', '[2]' ...\n- Higher card wins the prize (plus carry-over). Ties roll prize "
            f"into the pot for next round.\n- Highest total after 13 rounds wins."
        )

    def _start_round(self):
        gs = self.state.game_state
        gs["round"] += 1

        if gs["round"] > 13:
            s0, s1 = gs["player_scores"].values()
            if s0 > s1:     self.state.set_winner(0, f"P0 {s0} vs P1 {s1}")
            elif s1 > s0:   self.state.set_winner(1, f"P1 {s1} vs P0 {s0}")
            else:           self.state.set_draw(f"Both scored {s0}")
            return

        gs["current_prize"] = gs["prize_deck"][gs["round"] - 1]
        gs["pending_bids"] = {}
        gs["starting_player"] = 1 - gs["starting_player"]
        self.state.manually_set_current_player_id(gs["starting_player"])

        for pid in (0, 1):
            hand_str = " ".join(f"'[{self._val_to_face(c)}]'" for c in gs["player_hands"][pid])
            self.state.add_observation(to_id=pid, message=f"### Round {gs['round']}/13 - Prize: {self._val_to_face(gs['current_prize'])}  (worth {gs['current_prize'] + gs['carry_pot']})", observation_type=ta.ObservationType.GAME_MESSAGE)
            self.state.add_observation(to_id=pid, message=f"Your remaining hand: {hand_str}", observation_type=ta.ObservationType.GAME_BOARD)

    def step(self, action: str) -> Tuple[bool, Dict[str, Any]]:
        pid = self.state.current_player_id
        gs = self.state.game_state
        self.state.add_observation(from_id=pid, to_id=pid, message=action, observation_type=ta.ObservationType.PLAYER_ACTION)

        tokens = self.action_space.findall(action.lower())
        if len(tokens) != 1 or len(gs["pending_bids"]) >= 2: 
            self.state.set_invalid_move(reason="Action must contain exactly ONE bracketed card token.")
            return self.state.step()

        face = tokens[0]
        bid_val = self._face_to_val(face)
        if bid_val not in gs["player_hands"][pid]:
            self.state.set_invalid_move(reason="You no longer have that card.")
            return self.state.step()

        # record bid secretly
        gs["pending_bids"][pid] = bid_val
        gs["player_hands"][pid].remove(bid_val)

        # waiting for opponent?
        if len(gs["pending_bids"]) == 1: return self.state.step(rotate_player=True)

        bid0, bid1 = gs["pending_bids"][0], gs["pending_bids"][1]
        pot_value  = gs["current_prize"] + gs["carry_pot"]
        gs["carry_pot"] = 0

        reveal = (f"Bids: P0 {self._val_to_face(bid0)} vs P1 {self._val_to_face(bid1)} - ")
        if bid0 > bid1:     gs["player_scores"][0] += pot_value;    reveal += f"Player 0 wins {pot_value}."
        elif bid1 > bid0:   gs["player_scores"][1] += pot_value;    reveal += f"Player 1 wins {pot_value}."
        else:               gs["carry_pot"] += pot_value;           reveal += f"Tie -> pot now {gs['carry_pot']}."
        self.state.add_observation(message=reveal, observation_type=ta.ObservationType.GAME_MESSAGE)
        self.state.add_observation(message=f"Scores -> P0:{gs['player_scores'][0]}  P1:{gs['player_scores'][1]}", observation_type=ta.ObservationType.GAME_MESSAGE)

        self._start_round()
        return self.state.step(rotate_player=False)


from typing import Dict, Any

def create_board_str(game_state: Dict[str, Any]) -> str:
    """Render the gamemaster view of the 20 Questions game state."""
    
    target_word = game_state.get("target_word", "???")
    history = game_state.get("history", [])

    lines = []
    lines.append("🎯 Game: 20 Questions")
    lines.append("-" * 40)
    lines.append(f"Target Word: {target_word}")
    lines.append(f"Questions Asked: {len(history)} / 20")
    lines.append("-" * 40)

    if history:
        lines.append("📜 History:")
        for i, (q, a) in enumerate(history, start=1):
            lines.append(f"{i}. Q: {q}")
            lines.append(f"   A: {a}")
    else:
        lines.append("No questions asked yet.")

    return "\n".join(lines)



import os
import re, random, json, copy, importlib
from typing import Any, Dict, Optional, Tuple, Union

import textarena as ta
from textarena.envs.Crosswords.renderer import create_board_str


class CrosswordsEnv(ta.Env):
    def __init__(self, hardcore: Optional[bool]=False, max_turns: Optional[int]=100, num_words: Optional[int]=5):
        """
        Args:
            hardcore (Optional[bool]): Whether to use hardcore mode.
            max_turns (Optional[int]): Maximum number of turns allowed.
            num_words (Optional[int]): Number of words to use in the game.
        """
        super().__init__()
        self.hardcore = hardcore
        self.max_turns = max_turns
        self.num_words = num_words
        self._load_words(hardcore=hardcore)

    def get_board_str(self): return create_board_str(game_state=self.state.game_state)
    def _load_words(self, words_path: Optional[str]=None, hardcore: bool=False):
        try:
            if words_path is not None:
                if not os.path.exists(words_path): # Use provided path
                    raise FileNotFoundError(f"Words data file not found at: {words_path}")
                with open(words_path, "r", encoding="utf-8") as file:
                    word_data = file.readlines()
            else: # Use package resource
                with importlib.resources.files('textarena.envs.Crosswords').joinpath('words_clues.jsonl').open('r') as file:
                    word_data = file.readlines()
            self.word_data = [json.loads(x) for x in word_data if json.loads(x)["hardcore"] == hardcore]
            if not self.word_data:
                raise ValueError(f"No words found matching hardcore={hardcore} criteria.")
        except Exception as e:
            raise FileNotFoundError(f"Failed to load words data: {str(e)}")
    
    def reset(self, num_players: int, seed: Optional[int]=None):
        self.state = ta.SinglePlayerState(num_players=num_players, seed=seed, max_turns=self.max_turns) ## initialise the game_state
        game_board, placed_words, clues = self._generate_board() ## generate the game board and the placed words for the clues
        game_state={"solution": copy.copy(game_board), "board": self._hide_letters(game_board), "clues": clues, "placed_words": placed_words} # reset the state
        self.state.reset(game_state=game_state, player_prompt_function=self._generate_player_prompt)
        self._observer_current_board()

    def _generate_player_prompt(self, player_id: int, game_state: Dict[int, Any]) -> str:
        return (
            f"You are playing Crosswords.\nHere is the current state of the Crosswords grid. Each row and column are numbered.\n"
            "The cells that need to be populated with letters are represented by '_', and those that do not need words are represented by '.'.\n\n"
            "You can only provide one response per turn. Hence, plan your approach and risk appetite. Only guesses in the format of [row column letter] will be fetched from your response, e.g. [0 0 d], [1 2 G].\n"
            "As you play, the history of your choices will be appended below. Use the information to complete the game.\n"
        )

    def _observer_current_board(self):
        self.state.add_observation(f"Current Board:\n{self._render_board()}\nHere are the clues for the words you need to find:\n{self._clue_generator()}", observation_type=ta.ObservationType.GAME_BOARD)

    def _generate_board(self):
        ## init the sampled words, their directions and their clues
        sampled_word_data = random.sample(self.word_data, self.num_words)
        sampled_word_data_sorted = sorted(sampled_word_data, key=lambda x: len(x["word"]), reverse=True)
        words = [x["word"] for x in sampled_word_data_sorted]
        directions = {x["word"]: random.choice(["across", "down"]) for x in sampled_word_data_sorted}
        clues = {x["word"]: random.sample(list(x["clues"].values()), 1)[0] for x in sampled_word_data_sorted}

        ## generate the crossword grid
        grid_size = self._determine_initial_grid_size(words)
        grid = self._create_empty_grid(grid_size)

        placed_words = {}  # word: (row, col), where 0 is the starting index

        for word in words:
            placed = False
            if not placed_words:  # First word
                # Place the first word in the center of the grid
                if directions[word] == "across":
                    row = grid_size // 2
                    col = (grid_size - len(word)) // 2
                else:
                    row = (grid_size - len(word)) // 2
                    col = grid_size // 2

                if self._can_place_word(grid, word, directions[word], row, col):
                    self._place_word_on_grid(grid, word, directions[word], row, col)
                    placed_words[word] = (row, col, directions[word])
                    placed = True
            
            else:
                # Attempt to find overlaps
                possible_positions = self._find_overlaps(word, grid, placed_words, directions)
                random.shuffle(possible_positions)  # Randomize to add variability
                for pos in possible_positions:
                    row, col, direction = pos
                    if self._can_place_word(grid, word, direction, row, col):
                        self._place_word_on_grid(grid, word, direction, row, col)
                        placed_words[word] = (row, col, direction)
                        placed = True
                        break

            if not placed:
                # If no overlap placement is possible, try placing the word in any free position
                for row in range(grid_size):
                    for col in range(grid_size):
                        if self._can_place_word(grid, word, directions[word], row, col):
                            self._place_word_on_grid(grid, word, directions[word], row, col)
                            placed_words[word] = (row, col, directions[word])
                            placed = True
                            break
                    if placed:
                        break

        return grid, placed_words, clues

    def _determine_initial_grid_size(self, words):
        """ Determine the initial size of the grid based on the length of the longest word """
        max_length = max(len(word) for word in words)
        return round(max_length * 1.5)  # Ensures that the grid size is larger than the longest word to allow placement

    def _create_empty_grid(self, size):
        """ Create an empty grid of the specified size """
        return [["." for _ in range(size)] for _ in range(size)]

    def _can_place_word(self, grid, word, direction, row, col):
        """
        Check if a word can be placed on the grid at the specified position.

        Args:
            grid (List[List[str]]): The crossword grid.
            word (str): The word to be placed.
            direction (str): The direction in which the word is to be placed ("across" or "down").
            row (int): The row in which the word is to be placed.
            col (int): The column in which the word is to be placed.
        """
        if direction == "across":
            if col + len(word) > len(grid[0]):
                return False
            for i, letter in enumerate(word):
                current_cell = grid[row][col + i]
                if current_cell != "." and current_cell != letter:
                    return False
        else:  # "down"
            if row + len(word) > len(grid):
                return False
            for i, letter in enumerate(word):
                current_cell = grid[row + i][col]
                if current_cell != "." and current_cell != letter:
                    return False

        return True

    def _place_word_on_grid(self, grid, word, direction, row, col):
        """
        Place a word on the grid at the specified position.

        Args:
            grid (List[List[str]]): The crossword grid.
            word (str): The word to be placed.
            direction (str): The direction in which the word is to be placed ("across" or "down").
            row (int): The row in which the word is to be placed.
            col (int): The column in which the word is to be placed.
        """
        if direction == "across":
            for i, letter in enumerate(word):
                grid[row][col + i] = letter
        else:  # "down"
            for i, letter in enumerate(word):
                grid[row + i][col] = letter

    def _find_overlaps(self, word, grid, placed_words, directions):
        """
        Find all possible valid overlaps for the word with already placed words.
        
        Args:
            word (str): The word to be placed.
            grid (List[List[str]]): The crossword grid.
            placed_words (Dict[str, Tuple[int, int, str]]): A dictionary of placed words and their positions.
            directions (Dict[str, str]): A dictionary of words and their directions.

        Returns:
            List[Tuple[int, int, str]]: A list of possible overlaps in the format (row, col, direction
        """
        overlaps = []
        for placed_word, (p_row, p_col, p_direction) in placed_words.items():
            for i, letter in enumerate(word):
                for j, placed_letter in enumerate(placed_word):
                    if letter == placed_letter:
                        # Determine the possible position based on the direction of the placed word
                        if p_direction == 'across':
                            row = p_row - i
                            col = p_col + j
                            if directions[word] == 'down' and 0 <= row < len(grid) and 0 <= col < len(grid[0]):
                                if self._can_place_word(grid, word, 'down', row, col):
                                    overlaps.append((row, col, 'down'))
                        elif p_direction == 'down':
                            row = p_row + j
                            col = p_col - i
                            if directions[word] == 'across' and 0 <= row < len(grid) and 0 <= col < len(grid[0]):
                                if self._can_place_word(grid, word, 'across', row, col):
                                    overlaps.append((row, col, 'across'))
        return overlaps

    
    def _render_board(self):
        """ Print the grid for text display """
        ## should be C01, C03, ... C10, C11, ...
        header = "   " + " ".join(f"C{i:02}" for i in range(len(self.state.game_state['board'])))
        lines = [header]
        for i, row in enumerate(self.state.game_state['board']):
            ## should be R01, R02, ... R10, R11, ...
            row_str = f"R{i:02} "
            for j, val in enumerate(row): row_str += f" {val}  "
            lines.append(row_str)

        return "\n".join(lines)

    def _hide_letters(self, grid):
        """ Hide the letters in the grid """
        return [['_' if cell != "." else cell for cell in row] for row in grid]
    
    def step(self, action: str) -> Tuple[bool, ta.Info]:
        self.state.add_observation(from_id=self.state.current_player_id, message=action, observation_type=ta.ObservationType.PLAYER_ACTION) 
        ## validate the actions; note that the response can have multiple guesses at one go.
        matches = set(re.compile(r"\[(\d+)\s(\d+)\s([a-zA-Z])\]").findall(action)) # [row column letter]

        if not matches:
            self.state.set_invalid_move(reward=self._get_percentage_completion(), reason="The Player did not respond with valid 'row column letter'.")
        else:
            for match in matches:
                row, col, letter = match
                row, col, letter = int(row), int(col), str(letter)
                if row<0 or row>=len(self.state.game_state["board"]) or col<0 or col>=len(self.state.game_state["board"][0]):   self.state.set_invalid_move(reward=self._get_percentage_completion(), reason="The specified coordinate is out of bounds."); break
                elif self.state.game_state["board"][row][col] == ".":                                                           self.state.set_invalid_move(reward=self._get_percentage_completion(), reason=f"The specified coordinate is a black cell."); break
                elif not self.state.game_state["solution"][row][col].upper() == letter.upper():                                 self.state.set_invalid_move(reward=self._get_percentage_completion(), reason=f"Invalid move. The specified letter is incorrect."); break
                elif self.state.game_state["board"][row][col] != "_":                                                           self.state.set_invalid_move(reward=self._get_percentage_completion(), reason=f"The specified cell already contains a letter: {self.state.game_state['board'][row][col]}."); break
                else:           
                    self.state.game_state["board"][row][col] = letter.upper()
                    self._observer_current_board()
                    # self.state.add_observation(message=f"Board state: \n{self._render_board(self.state.game_state['board'], show_letters=True)}", observation_type=ta.ObservationType.GAME_BOARD)

            if self._is_game_over():                self.state.set_outcome(reward=1, reason=f"Congratulations! You completed the Crosswords puzzle.")
            elif self.state.check_turn_limit():     self.state.set_outcome(reward=self._get_percentage_completion(), reason=f"The turn limit has been reached. You completed {self._get_percentage_completion()*100} percent of the Crossword puzzle.")
        return self.state.step()

    def _get_percentage_completion(self) -> float:
        """ Compute the percentage of the crossword that has been solved so far """
        total_letter_cells = 0 # Count every cell that should eventually contain a letter
        filled_letter_cells = 0
        for row in self.state.game_state["board"]:
            for cell in row:
                if cell != ".": # not a black square
                    total_letter_cells += 1
                    if cell != "_" and cell.isalpha(): # already revealed / guessed
                        filled_letter_cells += 1
        if total_letter_cells == 0: # safety guard
            return 0.0
        return filled_letter_cells / total_letter_cells

    def _is_game_over(self) -> bool:
        """ Check if the game is over; Returns: (bool) True if the game is over, False otherwise """
        return all("_" not in row for row in self.state.game_state["board"])

    def _clue_generator(self, string_format=True):
        """ Generate a clue for a word; Returns: (str) The clue for the word. """
        res = []
        for i, set in enumerate(zip(self.state.game_state["placed_words"].values(), self.state.game_state["clues"].values())):
            res.append(f"{i+1}. {set[1]}: {set[0]}")
        return "\n".join(res) if string_format else res



from typing import Dict, Any, List, Tuple

def create_board_str(game_state: Dict[str, Any]) -> str:
    def draw_dice_row(dice: List[int]) -> Tuple[str, str, str]:
        tops = "   " + "  ".join("┌───┐" for _ in dice)
        faces = "   " + "  ".join(f"│ {d} │" for d in dice)
        bottoms = "   " + "  ".join("└───┘" for _ in dice)
        return tops, faces, bottoms

    def draw_player_box(player_id: int, dice: List[int]) -> str:
        top_border = f"┌─ PLAYER {player_id} ─{'─' * 26}┐"
        empty_line = "│" + " " * (len(top_border) - 2) + "│"
        tops, faces, bottoms = draw_dice_row(dice)
        return "\n".join([top_border, empty_line, f"│ {tops:<{len(top_border)-4}} │", f"│ {faces:<{len(top_border)-4}} │", f"│ {bottoms:<{len(top_border)-4}} │", empty_line, f"└{'─' * (len(top_border) - 2)}┘"])

    lines = []
    if len(game_state.get("remaining_dice", {})) <= 5:
        for pid in range(len(game_state.get("remaining_dice", {}))):
            if game_state.get("remaining_dice", {}).get(pid, 0) == 0: continue
            dice = game_state.get("dice_rolls", {}).get(pid, [])
            lines.append(draw_player_box(pid, dice))
            lines.append("")
    else: # Compact summary for > 5 players
        lines.append("Players & Dice:")
        for pid in range(len(game_state.get("remaining_dice", {}))):
            dice = game_state.get("dice_rolls", {}).get(pid, [])
            dice_str = " ".join(str(d) for d in dice)
            lines.append(f"  Player {pid} ({game_state.get('remaining_dice', {})[pid]} dice): {dice_str}")
        lines.append("")
    q = game_state.get("current_bid", {"quantity": 0, "face_value": 0}).get("quantity", 0)
    f = game_state.get("current_bid", {"quantity": 0, "face_value": 0}).get("face_value", 0)
    bid_box = f"""
┌─────────────────────────┐
│ Current Bid: {q} × face {f:<2}│
└─────────────────────────┘
""".strip()
    lines.append(bid_box)
    return "\n".join(lines)



"""Utils for Word Lists, common to language games. (by Dilan Hillier)"""

import importlib.resources
import re
from collections import defaultdict

# Load NLTK word list
import nltk
from nltk.corpus import words



def _parse_affix_rules(file_content: list[str]):
    # Regular expression pattern to match the affix rules
    pattern = re.compile(r"^(PFX|SFX)\s+(\w+)\s+(Y|N)\s+(\S+)\s+([\'|\w]+)\s+(\S+)$")

    # Lists to store the parsed rules
    prefixes = defaultdict(list)
    suffixes = defaultdict(list)

    # Iterate over each line in the file content
    for line in file_content:
        match = pattern.match(line.strip())
        if match:
            affix_type, flag, can_merge, strip_char, affix, condition = match.groups()
            rule = {
                "flag": flag,
                "can_merge": can_merge == "Y",
                "strip_char": strip_char if strip_char != "0" else None,
                "affix": affix,
                "condition": condition if condition != "." else None,
            }
            if affix_type == "PFX":
                prefixes[flag].append(rule)
            else:
                suffixes[flag].append(rule)

    return prefixes, suffixes


def _parse_condition(condition):
    """Parse the condition string in the affix rules."""
    if not condition:
        return None, None
    pattern = re.compile(r"(\[\^\S+\])?(\S+)?")
    match = pattern.match(condition)
    if match:
        exclude, include = match.groups()
        return exclude, include
    return None, None


def _split(line: str) -> tuple[str, str]:
    """Split a line from a .dic file into a word and its flags."""
    if "/" in line:
        return tuple(line.split("/"))
    return line.strip(), ""


class EnglishDictionary:
    """Dictionary Utils for English words."""

    def __init__(
        self, keep_proper_nouns=False, include_nltk=True, keep_non_alpha=False
    ):
        """Initialize the dictionary."""
        nltk.download("words")
        
        self.include_nltk = include_nltk
        self.keep_non_alpha = keep_non_alpha
        self.keep_proper_nouns = keep_proper_nouns
        self.uk_words = self._load_dic("en_GB.dic", "en.aff")
        # self.uk_words = self.expand(self.uk_words, self.uk_affs)
        self.us_words = self._load_dic("en_US.dic", "en.aff")
        # self.us_words = self.expand(self.us_words, self.us_affs)
        self.nltk_words = self._load_nltk(basic=False) if include_nltk else set()
        self.nltk_basic_words = self._load_nltk(basic=True) if include_nltk else set()

    def _filter(
        self,
        word_set: set[str],
    ) -> set[str]:
        filtered = set()
        for word in word_set:
            if word[0].isalpha() or self.keep_non_alpha:
                if word[0].islower() or self.keep_proper_nouns:
                    filtered.add(word)
        return filtered

    def _load_dic(
        self,
        filename: str,
        acc_filename: str,
    ) -> set[str]:
        """Load words from a .dic file inside the package's data folder."""
        with (
            importlib.resources.files("textarena.envs.utils.data")
            .joinpath(filename)
            .open("r") as f
        ):
            lines = f.readlines()[1:]  # Skip first line (word count)
        with (
            importlib.resources.files("textarena.envs.utils.data")
            .joinpath(acc_filename)
            .open("r") as f
        ):
            acc_lines = f.readlines()
        prefixes, suffixes = _parse_affix_rules(acc_lines)
        entries = set(_split(line) for line in lines)
        filtered = self._filter(entries)
        all_words = set()
        # first we add the base words
        for word, flag in filtered:
            all_words.add(word)
        prefixed_words = set()
        for word, flags in filtered:
            for flag in list(flags):
                if flag in prefixes:
                    # apply prefix
                    for rule in prefixes[flag]:
                        # check if the condition is met
                        exclude, include = _parse_condition(rule["condition"])
                        if exclude and any(
                            [word.endswith(exclude_char) for exclude_char in exclude]
                        ):
                            continue
                        if include and not word.endswith(include):
                            continue
                        if rule["strip_char"]:
                            if word.startswith(rule["strip_char"]):
                                new_word = (
                                    word[len(rule["strip_char"]) :] + rule["affix"]
                                )
                                all_words.add(new_word)
                                prefixed_words.add((new_word, flags))
                        else:
                            new_word = rule["affix"] + word
                            all_words.add(new_word)
                            prefixed_words.add(
                                (new_word, flags)
                            )  # may still need to apply suffixes
        for word, flags in filtered:
            for flag in list(flags):
                if flag in suffixes:
                    # apply suffix
                    for rule in suffixes[flag]:
                        # check if the condition is met
                        exclude, include = _parse_condition(rule["condition"])
                        if exclude and any(
                            [word.endswith(exclude_char) for exclude_char in exclude]
                        ):
                            continue
                        if include and not word.endswith(include):
                            continue
                        if rule["strip_char"]:
                            if word.endswith(rule["strip_char"]):
                                new_word = (
                                    word[: -len(rule["strip_char"])] + rule["affix"]
                                )
                                all_words.add(new_word)
                        else:
                            new_word = word + rule["affix"]
                            all_words.add(new_word)
        # finally we do merged prefixes and suffixes
        for word, flags in prefixed_words:
            for flag in list(flags):
                if flag in suffixes:
                    # apply suffix
                    for rule in suffixes[flag]:
                        # continue if flag is not mergeable
                        if not rule["can_merge"]:
                            continue
                        # check if the condition is met
                        exclude, include = _parse_condition(rule["condition"])
                        if exclude and word.endswith(exclude):
                            continue
                        if include and not word.endswith(include):
                            continue
                        if rule["strip_char"]:
                            if word.endswith(rule["strip_char"]):
                                new_word = (
                                    word[: -len(rule["strip_char"])] + rule["affix"]
                                )
                                all_words.add(new_word)
                        else:
                            new_word = word + rule["affix"]
                            all_words.add(new_word)
        all_words = self._filter(all_words)
        return all_words

    def _load_nltk(self, basic: bool) -> set[str]:
        nltk_words = set(words.words("en-basic") if basic else words.words("en"))
        return self._filter(nltk_words)

    def is_english_word(self, word: str) -> bool:
        """Check if a word is in the UK and/or US and/or NLTK English dictionary."""
        word = word.lower()
        return word in self.uk_words or word in self.us_words or word in self.nltk_words

    def get_all_words(self) -> set[str]:
        """Get all words in the dictionary as a set"""
        return self.uk_words | self.us_words | self.nltk_words

    def get_basic_words(self) -> set[str]:
        """Get all words in the basic NLTK dictionary as a set"""
        return self.nltk_basic_words



import re
from typing import Dict, Tuple, Optional, Any, List, Set

import textarena as ta


class CryptarithmEnv(ta.Env):
    _ACTION_RE = re.compile(r"\[\s*([A-Za-z])\s+(\d)\s*\]")

    def __init__(self, equation: str = "SEND + MORE = MONEY", max_turns: int = 100):
        """ equation : string of the form 'WORD [+ WORD …] = WORD' """
        super().__init__()
        self.equation_raw = equation.upper().replace(' ', '')
        lhs, rhs = self.equation_raw.split('=')
        self.addends: List[str] = lhs.split('+')
        self.result: str = rhs
        self.letters: Set[str] = set(''.join(self.addends) + self.result)
        self.first_letters: Set[str] = {w[0] for w in self.addends + [self.result]}
        self.max_turns = max_turns

        # mutable state (reset each episode)
        self.mapping: Dict[str, int] = {}      # current letter → digit
        self.digit_used: Dict[int, str] = {}   # digit → letter

    def reset(self, num_players: int, seed: Optional[int] = None):
        self.mapping.clear()
        self.digit_used.clear()
        self.state = ta.SinglePlayerState(num_players=num_players, max_turns=self.max_turns, seed=seed)
        self.state.reset(game_state={}, player_prompt_function=self._prompt)
        self._observe()

    def step(self, action: str) -> Tuple[bool, ta.Info]:
        self.state.add_observation(self.state.current_player_id, action, ta.ObservationType.PLAYER_ACTION)

        m = self._ACTION_RE.fullmatch(action.strip())
        if not m: self.state.set_invalid_move(self._progress(), "Bad action format. Use `[A 5]`."); return self.state.step()

        letter, digit = m.group(1).upper(), int(m.group(2))

        # basic validity checks
        if letter not in self.letters:                                  self.state.set_invalid_move(self._progress(), f"Letter {letter} not in puzzle.");                           return self.state.step()
        if digit in self.digit_used and self.digit_used[digit]!=letter: self.state.set_invalid_move(self._progress(), f"Digit {digit} already used by {self.digit_used[digit]}.");  return self.state.step()
        if letter in self.first_letters and digit == 0:                 self.state.set_invalid_move(self._progress(), "Leading digit of a word cannot be 0.");                      return self.state.step()

        # apply (re)assignment
        prev_digit = self.mapping.get(letter)
        if prev_digit is not None:
            del self.digit_used[prev_digit]
        self.mapping[letter] = digit
        self.digit_used[digit] = letter

        self._observe()

        # check win / max-turn
        if len(self.mapping) == len(self.letters):
            if self._equation_holds():      self.state.set_outcome(1.0, "Correct! Equation satisfied.")
            else:                           self.state.set_outcome(0.0, "Mapping complete but equation incorrect.")
        elif self.state.check_turn_limit(): self.state.set_outcome(self._progress(), "Move limit reached.")
        return self.state.step()

    def _prompt(self, player_id, game_state) -> str:    return "Map each letter to a unique digit so the arithmetic holds.\nAssign with `[A 5]`, re-assign anytime.\n"
    def _word_value(self, word: str) -> int:            return int(''.join(str(self.mapping[ch]) for ch in word))
    def _equation_holds(self) -> bool:                  return sum(self._word_value(w) for w in self.addends) == self._word_value(self.result) # leading-zero guard already enforced, so just compute integers
    def _progress(self) -> float:                       return len(self.mapping) / len(self.letters)

    def _render_board(self) -> str:
        # 1) Original equation with letters
        eq_letters = ' + '.join(self.addends) + f' = {self.result}'

        # 2) Partially-filled numeric view (digits for mapped letters, '_' for unknown)
        def show_word(w): return ''.join(str(self.mapping[ch]) if ch in self.mapping else '_' for ch in w)

        eq_digits = ' + '.join(show_word(w) for w in self.addends) + f' = {show_word(self.result)}'

        # 3) Current mapping table
        mapping_lines = ["Mapping:"]
        mapping_lines += [f"  {l} → {d}" for l, d in sorted(self.mapping.items())]
        if len(mapping_lines) == 1:
            mapping_lines.append("  (none yet)")

        # Combine everything
        return "\n"+'\n'.join([eq_letters, eq_digits, *mapping_lines])


    def _observe(self):
        self.state.add_observation(message=self._render_board()+f"\nAssigned: {len(self.mapping)}/{len(self.letters)} ({self._progress():.0%})", observation_type=ta.ObservationType.GAME_MESSAGE)



def create_board_str(game_state: dict) -> str:
    raise NotImplementedError
    lines = []
    phase_str = "🗣️ Communication Phase" if not game_state.get("is_decision_phase", False) else "🎯 Decision Phase"
    lines.append(f"╭── Iterated Prisoner's Dilemma ─────────────────────╮")
    lines.append(f"│ Round: {game_state.get('current_round', 1):<2} | Turn: {game_state.get('current_comm_turn', 0):<2} | Phase: {phase_str:<20}│")
    lines.append(f"╰────────────────────────────────────────────────────╯")
    lines.append("┌─ 📊 SCORES ─────────────────────┐")
    lines.append(f"│ Player 0: {game_state.get('scores', {}).get(0, 0):<3} pts               │")
    lines.append(f"│ Player 1: {game_state.get('scores', {}).get(1, 0):<3} pts               │")
    lines.append("└─────────────────────────────────┘")
    if game_state.get('history', []):
        lines.append("┌─ 📜 ROUND HISTORY ────────────────────────────────────┐")
        lines.append("│ Round │ Player 0     │ Player 1     │ Outcome         │")
        lines.append("├───────┼──────────────┼──────────────┼─────────────────┤")
        for round_info in game_state.get('history', []):
            d0 = round_info["decisions"].get(0, "?").capitalize()
            d1 = round_info["decisions"].get(1, "?").capitalize()
            if d0 == d1 == "Cooperate":                 outcome = "Both Cooperated"
            elif d0 == d1 == "Defect":                  outcome = "Both Defected"
            elif d0 == "Cooperate" and d1 == "Defect":  outcome = "P0 Sucker"
            elif d0 == "Defect" and d1 == "Cooperate":  outcome = "P1 Sucker"
            else:                                       outcome = "Unknown"
            lines.append(f"│ {round_info.get("round", "?"):^5} │ {d0:<12} │ {d1:<12} │ {outcome:<8} │")
        lines.append("└───────────────────────────────────────────────────────┘")
    return "\n".join(lines)



import random, re
from typing import Dict, Any, Optional, Tuple
import textarena as ta


class HighSocietyEnv(ta.Env):
    def __init__(self):
        super().__init__()
        self.money_cards = list(range(1, 12))   # 1-11
        self.action_space = re.compile(r"\[(11|10|[1-9])]", re.I)

    @staticmethod
    def _intlist_to_str(lst): return " ".join(f"'[{str(x)}]'" for x in sorted(lst))

    def reset(self, num_players: int = 2, seed: Optional[int] = None):
        self.state = ta.TwoPlayerState(num_players=num_players, seed=seed)
        deck = list(range(1, 11))
        random.shuffle(deck)
        game_state = {"round": 0, "prestige_deck": deck, "player_money": {0: self.money_cards.copy(), 1: self.money_cards.copy()}, "player_prestige": {0: 0, 1: 0}, "pending_bids": {}, "starting_player": 0}
        self.state.reset(game_state=game_state, player_prompt_function=self._prompt)
        self._next_auction()

    def _prompt(self, player_id: int, game_state: Dict[str, Any]) -> str:
        return (
            f"You are Player {player_id} in a game of HighSociety (2-player version).\n"
            f"Game flow:  Ten prestige cards are auctioned one after another.\n"
            f"Bidding:    Each auction, secretly choose a money card 1-11 and reveal.\n"
            f"            • Higher bid wins the prestige card and discards that money card.\n"
            f"            • Lower bid keeps their money card.\n"
            f"            • Tie -> both bids are returned and the same prestige card is re-auctioned.\n"
            f"Scoring:    After all ten auctions, add **remaining cash + prestige points**.\n"
            f"            Higher *net-worth* wins (exact tie -> draw).\n\n"
            f"**Action syntax**  →  bid a single card like '[7]' or '[11]'."
        )

    def _next_auction(self):
        gs = self.state.game_state
        gs["round"] += 1
        if not gs["prestige_deck"]: self._end_match(); return
        prize = gs["prestige_deck"].pop()
        gs["current_prize"] = prize
        gs["pending_bids"] = {}
        for pid in (0, 1):
            self.state.add_observation(to_id=pid, message=f"\n### Auction {gs['round']}/10  |  Prestige card: {prize}", observation_type=ta.ObservationType.GAME_MESSAGE)
            self.state.add_observation(to_id=pid, message=f"Your remaining money cards: {self._intlist_to_str(gs['player_money'][pid])}", observation_type=ta.ObservationType.GAME_BOARD)

    def step(self, action: str) -> Tuple[bool, Dict[str, Any]]:
        pid = self.state.current_player_id
        gs  = self.state.game_state
        self.state.add_observation(from_id=pid, to_id=pid, message=action, observation_type=ta.ObservationType.PLAYER_ACTION)
        t = self.action_space.findall(action)
        if len(t) != 1: self.state.set_invalid_move("Must include exactly one [X] money card."); return self.state.step()
        bid = int(t[0])
        if bid not in gs["player_money"][pid]: self.state.set_invalid_move("You no longer have that money card."); return self.state.step()
        gs["pending_bids"][pid] = bid
        if len(gs["pending_bids"]) == 1: return self.state.step(rotate_player=True)  # wait for opponent
        # both bids in
        bid0, bid1 = gs["pending_bids"][0], gs["pending_bids"][1]
        prize = gs["current_prize"]
        if bid0 > bid1:     winner, loser = 0, 1
        elif bid1 > bid0:   winner, loser = 1, 0
        else:  # tie -> redraw bids
            self.state.add_observation(message="Tie - bids returned. Rebid!", observation_type=ta.ObservationType.GAME_MESSAGE)
            gs["pending_bids"] = {}
            return self.state.step(rotate_player=True)
        gs["player_prestige"][winner] += prize
        gs["player_money"][winner].remove(gs["pending_bids"][winner])  # pay cost
        self.state.add_observation(message=f"P0 bid {bid0}, P1 bid {bid1}. Player {winner} wins prestige {prize} (total {gs['player_prestige'][winner]}).", observation_type=ta.ObservationType.GAME_MESSAGE)
        self._next_auction()
        return self.state.step(rotate_player=False)

    def _end_match(self):
        def networth(pid: int) -> int: return self.state.game_state["player_prestige"][pid]
        nw0, nw1 = networth(0), networth(1)
        if nw0 > nw1:   self.state.set_winner(0, f"Net-worth {nw0} > {nw1}")
        elif nw1 > nw0: self.state.set_winner(1, f"Net-worth {nw1} > {nw0}")
        else:           self.state.set_draw(f"Both net-worth {nw0} – exact tie.")



import re, random
from typing import Any, Dict, List, Tuple, Optional, Union

import textarena as ta
from textarena.envs.FifteenPuzzle.renderer import create_board_str

class FifteenPuzzleEnv(ta.Env):
    """ Fifteen Puzzle environment """
    def __init__(self, max_turns: int = 50):
        """ Initialize the Fifteen Puzzle environment """
        super().__init__()
        self.max_turns = max_turns

    def get_board_str(self):
        return create_board_str(game_state=self.state.game_state)
    
    def reset(self, num_players: int, seed: Optional[int] = None):
        """ Reset the environment to its initial state """
        self.state = ta.SinglePlayerState(num_players=num_players, seed=seed, max_turns=self.max_turns) ## initialize the game state
        self.board = self._generate_board() ## initialize the game state
        self.initial_board = [row[:] for row in self.board]  # Deep copy of the initial board
        game_state = {"board": self.board, "rendered_board": self._render_board(self.board)} ## reset the game state
        self.state.reset(game_state=game_state, player_prompt_function=self._generate_player_prompt)
        self._observe_current_state()  # Observe the initial state of the game

    def _observe_current_state(self) -> None:
        """Send current board and legal moves as observation."""
        r, c = self._get_empty_position()
        moves = {"[up]": r < 3, "[down]": r > 0, "[left]": c < 3, "[right]": c > 0}
        legal_moves = [m for m, valid in moves.items() if valid]
        msg = f"Current Board:\n\n{self.state.game_state['rendered_board']}\nAvailable Moves: {', '.join(legal_moves)}"
        self.state.add_observation(message=msg, observation_type=ta.ObservationType.GAME_BOARD)

    def _generate_player_prompt(self, player_id: int, game_state: Dict[int, Any]) -> str:
        return (
            f"You are Player {player_id}. You are playing the 15-Puzzle game.\n"
            "The objective of the game is to arrange the numbered tiles in ascending order from 1 to 15, with the empty space located in the bottom-right corner.\n"
            "To make a move, you can slide a tile into the empty space (represented by a double underscore, e.g. __) by using one of the following commands:\n"
            "- 'up': Move the tile below the empty space up.\n"
            "- 'down': Move the tile above the empty space down.\n"
            "- 'left': Move the tile to the right of the empty space left.\n"
            "- 'right': Move the tile to the left of the empty space right.\n"
            "To submit your move, type the direction (e.g., 'up', 'down', 'left', or 'right') in square brackets, e.g. [up].\n"
            "The current board layout is shown below. Use the information to solve the puzzle.\n"
        )
    
    def _generate_board(self):
        """ Generate a shuffled board configuration """
        tiles = list(range(1, 16)) + [None]
        random.shuffle(tiles)
        return [tiles[i:i + 4] for i in range(0, 16, 4)] # e.g. [[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12], [13, 14, 15, None]]
    
    def _render_board(self, board):
        """ Render the current board layout """
        rendered_board = ""
        for row in board:
            rendered_board += ' '.join(['__' if x is None else f"{x:2}" for x in row]) + "\n"
        return rendered_board
    
    def step(self, action: str) -> Tuple[bool, ta.Info]:
        """ Process the player's action and update the environment state """
        player_id = self.state.current_player_id
        self.state.add_observation(from_id=player_id, to_id=-1, message=action, observation_type=ta.ObservationType.PLAYER_ACTION) ## add the action to the game state
        action_search_pattern = re.compile(r"\[([a-zA-Z]+)\]") # e.g. [up]
        match = action_search_pattern.search(action)

        if match is None:
            reason=f"Invalid move format. Player {player_id} did not respond with a valid direction in square brackets."
            self.state.set_invalid_move(reward=self._get_percentage_completion(), reason=reason)

        else:
            direction = match.group(1)
            if not self._move(direction):
                reason=f"Invalid move. The tile cannot be moved in the specified direction."
                self.state.set_invalid_move(reward=self._get_percentage_completion(), reason=reason)

            else:
                self.state.game_state["rendered_board"] = self._render_board(self.board) ## update the rendered board
                message=f"Game Board:\n{self._render_board(self.board)}"
                self.state.add_observation(from_id=-1, to_id=player_id, message=message, observation_type=ta.ObservationType.GAME_BOARD)
            
        if self._is_solved(): ## check if the puzzle is solved
            reason=f"Congratulations! Player {player_id} have successfully solved the 15-Puzzle."
            self.state.set_winners(player_ids=[player_id], reason=reason)
        elif self.state.check_turn_limit():
            pct_completion = self._get_percentage_completion()
            reason=f"The turn limit has been reached. The model completed {pct_completion*100} percent of the puzzle"
            self.state.set_outcome(reward=pct_completion, reason=reason)
        self._observe_current_state()  # Observe the new state after the move
        return self.state.step()
    
    def _is_solved(self) -> bool:
        """ Check if the board is in a solved state """
        correct_tiles = list(range(1, 16)) + [None]
        current_tiles = [tile for row in self.board for tile in row]
        return current_tiles == correct_tiles

    def _move(self, direction: str) -> bool:
        """ Move a tile into the empty space if the direction is valid """
        empty_row, empty_col = self._get_empty_position()
        target_row, target_col = empty_row, empty_col

        if direction == 'up' and empty_row < 3:         target_row += 1
        elif direction == 'down' and empty_row > 0:     target_row -= 1
        elif direction == 'left' and empty_col < 3:     target_col += 1
        elif direction == 'right' and empty_col > 0:    target_col -= 1
        else:                                           return False ## invalid move

        ## swap the target tile with the empty tile
        self.board[empty_row][empty_col], self.board[target_row][target_col] = (self.board[target_row][target_col], self.board[empty_row][empty_col])
        return True
    
    def _get_empty_position(self):
        for r in range(4):
            for c in range(4):
                if self.board[r][c] is None:
                    return r, c

    def _get_percentage_completion(self) -> float:
        goal = list(range(1, 16)) + [None]
        correct = 0
        total = 0
        # Flatten all 3 boards for easier comparison
        flat_current = [tile for row in self.board for tile in row]
        flat_initial = [tile for row in self.initial_board for tile in row]
        for idx, goal_tile in enumerate(goal):
            if flat_initial[idx] == goal_tile: continue  # Skip tiles that were already in the right place initially
            total += 1
            if flat_current[idx] == goal_tile:
                correct += 1
        return correct / total if total > 0 else 0.0



import re
from collections import deque
from typing import Optional, Dict, Tuple, List, Any

import textarena as ta
from textarena.envs.SimpleTak.renderer import create_board_str

class SimpleTakEnv(ta.Env):
    def __init__(self, board_size: int = 5):
        """
        Args:
            board_size (int): The size of the NxN board (default 5).
        """
        super().__init__()
        self.board_size = board_size
        self.cell_mapping = {i: (i // board_size, i % board_size) for i in range(board_size * board_size)}

    def get_board_str(self): return create_board_str(board=self.state.game_state["board"], board_size=self.board_size)
    def reset(self, num_players: int = 2, seed: Optional[int] = None):
        self.state = ta.TwoPlayerState(num_players=num_players, seed=seed)
        self.state.reset(game_state={"board": [['' for _ in range(self.board_size)] for _ in range(self.board_size)]}, player_prompt_function=self._prompt)
        self._observe_current_state()

    def _prompt(self, player_id: int, game_state: Dict[str, Any]) -> str:
        return (
            f"You are Player {player_id} in SimpleTak.\n"
            f"On the board, your stones appear as '{'O' if player_id == 0 else 'X'}' and "
            f"your opponent's stones appear as '{'O' if player_id == 1 else 'X'}'.\n\n"
            "On your turn, choose one empty cell (by its numbered index) and place your stone there.\n"
            "For example, '[12]' places your stone in cell 12.\n\n"
            "Your objective is to form a continuous path of your stones that connects two opposite edges of the board "
            "(top-to-bottom or left-to-right)."
        )

    def _observe_current_state(self) -> None:
        available_moves = []
        for i in range(self.board_size * self.board_size):
            r, c = self.cell_mapping[i]
            if self.state.game_state["board"][r][c] == '': 
                available_moves.append(f"[{i}]")
        self.state.add_observation(message=f"Current Board:\n\n{self._render_board()}\nAvailable Moves: " + ", ".join(available_moves), observation_type=ta.ObservationType.GAME_BOARD)

    def _render_board(self) -> str:
        max_cell_num = self.board_size * self.board_size - 1
        digit_count = len(str(max_cell_num))
        cell_width = max(digit_count, 2)  # at least 2 for occupant symbols
        def cell_str(r: int, c: int) -> str:
            if self.state.game_state["board"][r][c] == '': return str(r * self.board_size + c) # If empty, show cell number
            else: return self.state.game_state["board"][r][c] # Occupied by 'O' or 'X'
        def build_hline() -> str:
            line_parts = []
            for _ in range(self.board_size): line_parts.append("-" * (cell_width + 2))  # +2 for spacing around content
            return "+" + "+".join(line_parts) + "+"
        lines = []
        lines.append(build_hline())
        for r in range(self.board_size):
            row_cells = []
            for c in range(self.board_size):
                text = cell_str(r, c)
                text_centered = f" {text:^{cell_width}} "
                row_cells.append(text_centered)
            row_line = "|" + "|".join(row_cells) + "|"
            lines.append(row_line)
            lines.append(build_hline())
        return "\n".join(lines)

    def step(self, action: str) -> Tuple[bool, ta.Info]:
        symbol = 'O' if self.state.current_player_id == 0 else 'X'
        self.state.add_observation(from_id=self.state.current_player_id, message=action, observation_type=ta.ObservationType.PLAYER_ACTION)
        match = re.compile(r"\[\s*(\d+)\s*\]").search(action) # Regex to parse moves like [12]
        if match is None:
            self.state.set_invalid_move(reason="Invalid move format")
        else:
            cell_num = int(match.group(1))
            if cell_num not in self.cell_mapping: # Check if cell_num in valid range
                self.state.set_invalid_move(reason=f"Invalid cell number {cell_num}. Must be between 0 and {self.board_size**2 - 1}.")
            else:
                row, col = self.cell_mapping[cell_num]
                board = self.state.game_state["board"]
                if board[row][col] == '':
                    board[row][col] = symbol # Place the stone
                    self.state.add_observation(message=f"Player {self.state.current_player_id} placed their symbol ({symbol}) in cell {cell_num}.", observation_type=ta.ObservationType.GAME_ACTION_DESCRIPTION)
                    if self._check_win(symbol): # Check for a winning path
                        self.state.set_winner(player_id=self.state.current_player_id, reason=f"Player {self.state.current_player_id} ('{symbol}') connected two opposite edges!")
                    else:
                        if all(board[r][c] != '' for r in range(self.board_size) for c in range(self.board_size)): # If board is fully occupied and no winner => draw
                            self.state.set_draw(reason="The board is full. It's a draw!")
                else:
                    self.state.set_invalid_move(reason=f"Cell {cell_num} is already occupied. Choose an empty cell.")
        self._observe_current_state()
        return self.state.step()

    def _check_win(self, symbol: str) -> bool:
        n   = self.board_size
        bd  = self.state.game_state["board"]
        dirs = [(0,1), (1,0), (0,-1), (-1,0)]           # 4-neighbour connectivity

        def bfs(starts: List[Tuple[int,int]], target_edge) -> bool:
            """ Generic flood-fill. `target_edge` is a lambda that tests whether (r,c) lies on the opposite edge we’re trying to reach. """
            q = deque(starts)
            vis = set(starts)
            while q:
                r, c = q.popleft()
                if target_edge(r, c):
                    return True
                for dr, dc in dirs:
                    nr, nc = r+dr, c+dc
                    if 0 <= nr < n and 0 <= nc < n and (nr, nc) not in vis and bd[nr][nc] == symbol:
                        vis.add((nr, nc))
                        q.append((nr, nc))
            return False

        # top → bottom
        top_starts = [(0, c) for c in range(n) if bd[0][c] == symbol]
        if bfs(top_starts, lambda r, _c: r == n-1):
            return True

        # left → right
        left_starts = [(r, 0) for r in range(n) if bd[r][0] == symbol]
        if bfs(left_starts, lambda _r, c: c == n-1):
            return True

        return False


import re, nltk, random 
from nltk import pos_tag
from nltk.corpus import words
from typing import Optional, Tuple, List, Dict, Any

import textarena as ta
from textarena.envs.Wordle.renderer import create_board_str
from textarena.envs.utils.word_lists import EnglishDictionary

try:
    pos_tag(['test'])
except LookupError:
    nltk.download('averaged_perceptron_tagger_eng', quiet=True)

try:
    words.words()
except LookupError:
    nltk.download('words', quiet=True)
    
class WordleEnv(ta.Env):
    def __init__(self, word_length: int = 5, num_guesses: int = 6, hardcore: Optional[bool] = False):
        """ Initializes the Wordle environment """
        super().__init__()
        self.word_length = word_length
        self.num_guesses = num_guesses
        self._load_word_list(hardcore=hardcore)
        self.dictionary = EnglishDictionary(keep_proper_nouns=False, include_nltk=True)

    def get_board_str(self):
        return create_board_str(game_state=self.state.game_state)
        
    def _check_word(self, word: str) -> bool:
        return self.dictionary.is_english_word(word)
    
    def _load_word_list(self, hardcore: bool = False) -> None:
        """ Load the word list based on the 'hardcore' parameter """
        word_list = words.words("en") if hardcore else words.words("en-basic") # Get word list
        self.word_list = [word for word in word_list if pos_tag([word])[0][1] in ["NN"] and len(word) == self.word_length] # Filter words based on POS tags

    def reset(self, num_players: int = 1, seed: Optional[int] = None):
        self.state = ta.SinglePlayerState(num_players=num_players, seed=seed)
        game_state = {"secret_word": random.choice(self.word_list), "guess_history": [], "word_length": self.word_length, "num_guesses": self.num_guesses}
        self.state.reset(game_state=game_state, player_prompt_function=self._generate_player_prompt)
    
    def _generate_player_prompt(self, player_id: int, game_state: Dict[int, Any]) -> str:
        return (
            f"You are Playing Wordle.\nA secret {game_state['word_length']}-letter word has been chosen. You have {game_state['num_guesses']} attempts to guess it.\n"
            "For each guess, wrap your word in square brackets (e.g., '[apple]').\nFeedback for each letter will be given as follows:\n"
            "  - G (green): correct letter in the correct position\n"
            "  - Y (yellow): letter exists in the word but in the wrong position\n"
            "  - X (wrong): letter is not in the word\n"
            "Enter your guess to begin.\n"
        )

    def step(self, action: str) -> Tuple[bool, ta.Info]:
        player_id = self.state.current_player_id
        self.state.add_observation(message=action, observation_type=ta.ObservationType.PLAYER_ACTION)
        match = re.search(r"\[(\w+)\]", action) # Extract the guess using regex

        if match is None:
            self.state.set_invalid_move(reward=self._get_percentage_completion(), reason=f"You tried submitting a word in the wrong format. Please make sure to use squared brackets.")
            return self.state.step()
        
        word = match.group(1).lower()
        if len(word) != self.state.game_state["word_length"]:
            self.state.set_invalid_move(reward=self._get_percentage_completion(), reason=f"Your word must be exactly {self.state.game_state['word_length']} letters.")
            return self.state.step()
        
        # Check if the word has been guessed before
        previous_words = [guess_word for guess_word, _ in self.state.game_state["guess_history"]]
        if word in previous_words:
            self.state.set_invalid_move(reward=self._get_percentage_completion(), reason=f"You have already guessed '{word}' before. Please try a different word.")
            return self.state.step()
        
        if not self._check_word(word):
            self.state.set_invalid_move(reward=self._get_percentage_completion(), reason=f"'{word}' is not an English word.")
            return self.state.step()

        
        feedback = self._evaluate_guess(word) # Evaluate the word
        self.state.game_state["guess_history"].append((word, feedback)) # Save the guess and feedback

        # Update board views
        self.state.game_state["rendered_board"] = self._render_board()
        self.state.game_state["player_view"] = self._render_player_view(player_id)

        # Check for win condition (all letters green)
        if all(f == "G" for f in feedback):
            self.state.set_outcome(reward=1, reason=f"Congratulations! You guessed the word correctly!")
        else:
            self.state.add_observation(message=f"You submitted [{word}].\nFeedback:\n{self._render_player_view(player_id)}\nYou have {self.state.game_state['num_guesses'] - self.state.turn - 1} guesses left.", observation_type=ta.ObservationType.GAME_MESSAGE)

        # check if max num guesses reached
        if len(self.state.game_state["guess_history"]) >= self.num_guesses and not self.state.done:
            pct_complete = self._get_percentage_completion()
            secret = self.state.game_state["secret_word"]
            reason = f"The turn limit has been reached. You didn't guess the word, but your best guess matched {round(pct_complete * 100)}% of the letters in the correct positions.\nThe secret word was: **{self.state.game_state['secret_word']}**."
            self.state.set_outcome(reward=pct_complete, reason=reason)

        return self.state.step()

    def _evaluate_guess(self, guess: str) -> List[str]:
        """
        Evaluates the player's guess against the secret word and returns feedback for each letter.
        
        Feedback:
            - "green": correct letter in the correct position.
            - "yellow": letter is in the word but in the wrong position.
            - "wrong": letter is not in the word.
        
        Args:
            guess (str): The player's guess.
        
        Returns:
            List[str]: A list of feedback tokens for each letter.
        """
        feedback = [None] * self.state.game_state["word_length"]
        secret_list = list(self.state.game_state["secret_word"])
        guess_list = list(guess)

        # First pass: mark correct letters in the correct position (green)
        for i in range(self.word_length):
            if guess_list[i] == secret_list[i]:
                feedback[i] = "G"
                secret_list[i] = None  # Mark this letter as accounted for

        # Second pass: mark correct letters in the wrong position (yellow) or wrong letters
        for i in range(self.word_length):
            if feedback[i] is None:
                if guess_list[i] in secret_list:
                    feedback[i] = "Y"
                    # Remove the first occurrence of guess_list[i] from secret_list
                    index = secret_list.index(guess_list[i])
                    secret_list[index] = None
                else:
                    feedback[i] = "X"
        return feedback
    
    def _render_board(self) -> str:
        """ Renders the board in full Wordle format. """
        history = self.state.game_state["guess_history"]
        if not history:
            return "No guesses yet."

        output = []
        for word, feedback in history:
            letters_row = "| Letter  | " + " ".join(word.upper()) + " |"
            divider_row = "|---------|" + "--" * self.state.game_state['word_length'] + "--"
            status_row = "| Status  | " + " ".join(feedback) + " |"
            output.append(f"{letters_row}\n{divider_row}\n{status_row}\n")

        return "\n".join(output)
    
    def _render_player_view(self, player_id: int) -> str:
        """ Renders a simplified player view (letters and feedback only). """
        if not self.state.game_state["guess_history"]:
            return "No guesses yet."
        
        # Get the most recent guess
        word, feedback = self.state.game_state["guess_history"][-1]
        word_row = " ".join(word.upper())
        feedback_row = " ".join(feedback)
        return f"{word_row}\n{feedback_row}"

    def _get_percentage_completion(self) -> float:
        """ 
        Compute completion based on the most recent guess for RL training.
        This encourages the model to maximize immediate performance.
        Returns a float ∈ [0.0, 1.0]
        """
        if not self.state.game_state.get("guess_history", []):
            return 0.0
        
        # Get the most recent guess feedback
        _, latest_feedback = self.state.game_state["guess_history"][-1]
        
        # Calculate the percentage of letters that are green (correct position) and yellow (correct letter, wrong position)
        greens = sum(1 for f in latest_feedback if f == "G")
        yellows = sum(1 for f in latest_feedback if f == "Y") * 0.5  # Yellow letters count as half for completion

        # Calculate the percentage completion based on the number of green and yellow letters
        return (greens + yellows) / self.state.game_state["word_length"]
    


import random, re
from typing import Dict, Any, Optional, Tuple
import textarena as ta

# TODO work in progress
class LeTrucEnv(ta.Env):
    """
    Minimal 2-player 'Le Truc'.
    • Deck = 32 cards (remove 8s/9s/10s). Rank order: 3 2 A K Q J 7 6 5 4.
    • Each gets 3 cards. Trick winner is highest card of led suit.
    • At any time before a trick result a player may [raise] to increase hand value by +1. Opponent may [accept] or [fold].
    • First to 12 match-points wins (or earlier if someone folds).
    """
    order = ["3", "2", "A", "K", "Q", "J", "7", "6", "5", "4"]

    def __init__(self):
        super().__init__()
        self.action_space = re.compile(
            r"""\[
                (?P<verb>play|raise|accept|fold)            # action keyword
                (?:\s+(?P<card>(10|[234567JQKA])))?         # optional rank
            \]""",
            re.IGNORECASE | re.VERBOSE,
        )
        # build the 32-card deck
        suits = "♣♦♥♠"
        self.deck = [r + s for r in self.order for s in suits]

    def reset(self, num_players: int = 2, seed: Optional[int] = None):
        self.state = ta.TwoPlayerState(num_players=num_players, seed=seed)
        self.state.reset(game_state={"match_points": {0: 0, 1: 0}, "hand_points": 1}, player_prompt_function=self._prompt)
        self._deal_hand()

    def _prompt(self, player_id: int, game_state: Dict[str, Any]) -> str:
        return "a summary of the game rules"
    

    def _deal_hand(self):
        gs = self.state.game_state
        d = self.deck.copy()
        random.shuffle(d)
        gs["hands"] = {0: d[:3], 1: d[3:6]}
        gs["tricks"] = []
        gs["led_card"] = None
        gs["raiser"] = None
        self.state.manually_set_current_player_id(0)

        for pid in (0, 1):
            self.state.add_observation(to_id=pid, message=f"### New hand worth {gs['hand_points']} pt(s)\nYour cards: {' '.join(gs['hands'][pid])}", observation_type=ta.ObservationType.GAME_BOARD)

    def _rank_idx(self, card: str) -> int: return self.order.index(card[:-1])

    def step(self, action: str) -> Tuple[bool, Dict[str, Any]]:
        pid = self.state.current_player_id
        gs  = self.state.game_state
        self.state.add_observation(from_id=pid, message=action, observation_type=ta.ObservationType.PLAYER_ACTION)

        m = self.action_space.search(action)
        if not m: self.state.set_invalid_move(reason="Unrecognised action."); return self.state.step()
        verb  = m.group("verb").lower()
        if verb == "raise":
            if gs.get("raiser") is not None:
                self.state.set_invalid_move("Already raised.")
                return self.state.step()

            gs["hand_points"] += 1
            gs["raiser"] = pid
            self.state.current_player_id = 1 - pid
            self.state.add_observation(message=f"P{pid} raises - hand now {gs['hand_points']} pts. Opponent must '[accept]' or '[fold]'.", observation_type=ta.ObservationType.GAME_MESSAGE)
            return self.state.step()

        if verb == "accept":
            if gs.get("raiser") is None or pid == gs["raiser"]:
                self.state.set_invalid_move(reason="No raise to accept.")
                return self.state.step()

            gs["raiser"] = None
            # resume with whoever still has to follow suit first
            self.state.current_player_id = gs.get("led_to", pid)
            return self.state.step()

        if verb == "fold":
            if gs.get("raiser") is None or pid == gs["raiser"]:
                self.state.set_invalid_move(reason="Cannot fold now.")
                return self.state.step()

            winner = 1 - pid
            gs["match_points"][winner] += gs["hand_points"]
            self.state.set_winner(winner, reason="Opponent folded.")
            return self.state.step()

        if verb == "play":
            rank = m.group("card").upper()

            # ensure player owns that rank (ignore suits in input)
            if rank not in [c[:-1] for c in gs["hands"][pid]]:
                self.state.set_invalid_move(reason="You don't hold that rank (suits ignored in input).")
                return self.state.step()

            # pick the first matching suit in hand
            idx = next(i for i, c in enumerate(gs["hands"][pid]) if c.startswith(rank))
            card_str = gs["hands"][pid].pop(idx)

            if gs["led_card"] is None:
                # lead the trick
                gs["led_card"] = (pid, card_str)
                gs["led_to"]   = 1 - pid
                self.state.current_player_id = 1 - pid
                self.state.add_observation(message=f"P{pid} leads {card_str}.", observation_type=ta.ObservationType.GAME_MESSAGE)
            else:
                # follow, decide trick winner
                lead_pid, lead_card = gs["led_card"]
                win_pid = pid if self._rank_idx(card_str) > self._rank_idx(lead_card) else lead_pid

                gs["tricks"].append(win_pid)
                self.state.add_observation(message=f"P{pid} plays {card_str}. Trick to P{win_pid}.", observation_type=ta.ObservationType.GAME_MESSAGE)

                gs["led_card"] = None
                if len(gs["tricks"]) == 3:                                  # hand over
                    winner = 0 if gs["tricks"].count(0) > 1 else 1
                    gs["match_points"][winner] += gs["hand_points"]
                    if gs["match_points"][winner] >= 12:                    # match over
                        self.state.set_winner(winner, "Reached 12 points.")
                        return self.state.step()

                    # new hand
                    gs["hand_points"] = 1
                    self._deal_hand()
                    return self.state.step(rotate_player=False)
                self.state.current_player_id = win_pid
            return self.state.step(rotate_player=False)
        # should never reach here
        self.state.set_invalid_move("Unrecognised action.")
        return self.state.step()



import json, logging
import ssl
import asyncio
from typing import List, Optional, Tuple, Dict, Any, Union
from urllib.parse import urlencode
import warnings
from urllib3.exceptions import InsecureRequestWarning
import uuid
from textarena.envs.registration import ENV_REGISTRY


# online play specific imports
try:
    import requests, websockets
except ImportError:
    raise ImportError("'requests' and 'websockets' libraries are required for online play. Install them with: 'pip install textarena[online]' OR pip install requests, websockets'")


# Suppress SSL warnings
warnings.filterwarnings('ignore', category=InsecureRequestWarning)

# Server URLs - Change these to match your server
MATCHMAKING_WS_URI = "wss://matchmaking.textarena.ai/ws"
MATCHMAKING_HTTP_URI = "https://matchmaking.textarena.ai"

# Environment ID mapping
NAME_TO_ID_DICT = { # TODO update with more games that map to Supabase
    "Chess-v0": 0,
    "ConnectFour-v0": 1,
    "DontSayIt-v0": 3,
    "Battleship-v0": 5,
    "LiarsDice-v0": 6,
    "SimpleNegotiation-v0": 8,
    "Poker-v0": 9,
    "SpellingBee-v0": 10,
    "SpiteAndMalice-v0": 11,
    "Stratego-v0": 12,
    "Tak-v0": 13,
    "TruthAndDeception-v0": 14,
    "UltimateTicTacToe-v0": 15,
    "WordChains-v0": 16,
    "TicTacToe-v0": 35,
    "Breakthrough-v0": 37,
    "Checkers-v0": 38,
    "KuhnPoker-v0": 46,
    "LetterAuction-v0": 47,
    "MemoryGame-v0": 48,
    "Nim-v0": 50,
    "Othello-v0": 51,
    "PigDice-v0": 52,
    "SimpleBlindAuction-v0": 56,
    "Snake-v0": 69,
    "SecretMafia-v0": 75,
    "WildTicTacToe-v0": 77,
    "ReverseTicTacToe-v0": 78,
    "RandomizedTicTacToe-v0": 79,
    "QuantumTicTacToe-v0": 80,
    "IteratedRockPaperScissors-v0": 81,
    "Subset-v0": [0, 3, 35, 51, 52, 63, 66, 67, 68, 70, 75, 81]
}

def strip_env_variant(env_id: str) -> str:
    for suffix in ["-train", "-raw"]:
        if env_id.endswith(suffix):
            return env_id[: -len(suffix)]
    return env_id

class DynamicWrapperProxy:
    """A proxy that dynamically applies wrappers once the environment is known."""
    
    def __init__(self, base_env, env_id_to_wrappers_map):
        self.base_env = base_env
        self.env_id_to_wrappers_map = env_id_to_wrappers_map
        self.wrapped_env = None
        self.matched_env_id = None
        self._wrappers_applied = False
        
    def _apply_wrappers_for_env(self, env_id):
        """Apply the appropriate wrappers for the given environment ID."""
        if self._wrappers_applied:
            return  # Already wrapped
            
        self.matched_env_id = env_id
        
        # Find the wrappers for this environment
        wrappers = self.env_id_to_wrappers_map.get(env_id, [])
        
        if wrappers:
            print(f"[DynamicWrapper] Applying wrappers for matched environment '{env_id}':")
            self.wrapped_env = self.base_env
            for wrapper in wrappers:
                print(f"  - {wrapper.__name__ if hasattr(wrapper, '__name__') else wrapper}")
                self.wrapped_env = wrapper(self.wrapped_env)
        else:
            # print(f"[DynamicWrapper] No wrappers found for '{env_id}', using base environment")
            self.wrapped_env = self.base_env
        
        self._wrappers_applied = True
        # print(f"[DynamicWrapper] Wrappers applied successfully for '{env_id}'")
    
    def _get_active_env(self):
        """Get the currently active environment (wrapped or base)."""
        # Check if we need to apply wrappers based on matched environment
        if not self._wrappers_applied and hasattr(self.base_env, 'matched_env_name') and self.base_env.matched_env_name:
            # print(f"[DynamicWrapper] Detected matched environment: {self.base_env.matched_env_name}")
            self._apply_wrappers_for_env(self.base_env.matched_env_name)
        
        return self.wrapped_env if self.wrapped_env is not None else self.base_env
    
    def get_observation(self):
        """Special handling for get_observation to ensure wrappers are applied."""
        # Check if we need to apply wrappers
        if not self._wrappers_applied:
            # print(f"[DynamicWrapper] get_observation called, checking for matched environment...")
            if hasattr(self.base_env, 'matched_env_name') and self.base_env.matched_env_name:
                # print(f"[DynamicWrapper] Found matched environment: {self.base_env.matched_env_name}")
                self._apply_wrappers_for_env(self.base_env.matched_env_name)
            # else:
                # print(f"[DynamicWrapper] No matched environment found yet")
        
        active_env = self._get_active_env()
        return active_env.get_observation()
    
    def step(self, action):
        """Special handling for step to ensure wrappers are applied."""
        active_env = self._get_active_env()
        return active_env.step(action)
    
    def reset(self, *args, **kwargs):
        """Special handling for reset to ensure wrappers are applied."""
        # Reset the base environment first
        result = self.base_env.reset(*args, **kwargs)
        
        # Check if we now have a matched environment and apply wrappers
        if hasattr(self.base_env, 'matched_env_name') and self.base_env.matched_env_name and not self._wrappers_applied:
            # print(f"[DynamicWrapper] After reset, applying wrappers for: {self.base_env.matched_env_name}")
            self._apply_wrappers_for_env(self.base_env.matched_env_name)
            
        return result
    
    def close(self):
        """Special handling for close."""
        active_env = self._get_active_env()
        return active_env.close()
    
    def __getattr__(self, name):
        """Delegate all other attribute access to the active environment."""
        return getattr(self._get_active_env(), name)


class OnlineEnvWrapper:
    def __init__(self, env_ids: List[int], env_id_names: List[str], model_name: str, model_token: str):
        self.env_ids = env_ids
        self.env_id_names = env_id_names  # Store the original env_id names
        self.model_name = model_name
        self.model_token = model_token
        
        # Connection variables
        self.websocket = None
        self.matchmaking_websocket = None
        self.game_url = None
        self.environment_id = None
        self.env_id = None
        self.matched_env_name = None  # Store the matched environment name
        
        # Create mapping from env_id names to wrappers
        self.env_id_to_wrappers_map = {}
        for env_name in env_id_names:
            if env_name in ENV_REGISTRY:
                env_spec = ENV_REGISTRY[env_name]
                self.env_id_to_wrappers_map[env_name] = env_spec.default_wrappers or []
        
        # The full observations are stored as a dictionary mapping player id -> list of (sender_id, message) tuples
        self.full_observations = {}
        
        # Game state tracking
        self.current_player_id = None
        self.current_observation = None
        self.game_over = False
        self.server_shutdown = False
        self.game_over_timeout = 30.0

        self.rewards = {}
        self.step_info = {}
        self.game_info = {}
        
        # Timeouts
        self.matchmaking_timeout = 1800

        
        # Async queues for incoming/outgoing messages
        self.message_queue = asyncio.Queue()
        self.action_queue = asyncio.Queue()
        self.matchmaking_queue = asyncio.Queue()
        
        # State tracking
        self.in_game = False
        self.pending_action = False
        self.update_task = None
        self.matchmaking_complete = False
        
        # For compatibility
        DummyState = type("DummyState", (), {})
        self.state = DummyState()
        self.state.role_mapping = {0: "Player 0", 1: "Player 1", -1: "GAME"}

    async def _message_receiver(self):
        """
        Background task that listens to messages from the game server websocket
        and places them into the internal message queue for processing.
        
        Also performs a quick check for 'server shutdown' messages to gracefully exit early.
        """
        try:
            while True:
                try:
                    message = await self.websocket.recv()
                    print(f"Received: {message}")

                    # put the raw message into the queue for processing
                    await self.message_queue.put(message)
                    
                    # Proactively check for 'server_shutdown' command to allow early exit
                    try:
                        msg_data = json.loads(message)
                        if msg_data.get("command") == "server_shutdown":
                            print("Server shutdown message detected in receiver")
                            self.server_shutdown = True
                    except:
                        pass
                        
                except websockets.exceptions.ConnectionClosed:
                    print("WebSocket connection closed by server")
                    self.server_shutdown = True
                    break

                except Exception as e:
                    print(f"Error receiving message: {e}")
                    break

        except Exception as e:
            print(f"Message receiver error: {e}")
            self.server_shutdown = True

    async def _matchmaking_receiver(self):
        """
        Background task that listens to the matchmaking websocket.
        
        It reads and queues all messages until a match is found or the connection is closed.
        """
        try:
            while not self.matchmaking_complete:
                try:
                    message = await self.matchmaking_websocket.recv()
                    print(f"Received from matchmaking: {message}")

                    # pass the raw message to the matchmaking queue for processing
                    await self.matchmaking_queue.put(message)

                except websockets.exceptions.ConnectionClosed:
                    print("Matchmaking WebSocket connection closed")
                    break

                except Exception as e:
                    print(f"Error receiving matchmaking message: {e}")
                    break

        except Exception as e:
            print(f"Matchmaking receiver error: {e}")

    async def _action_sender(self):
        """
        Background task that listens for actions from the action_queue and sends them to the game server.

        Waits for actions like `"play x y"` or `"bet 3"`, and handles graceful shutdown when it receives `"CLOSE"`.
        """
        try:
            while True:
                # Wait for the next action to send
                action = await self.action_queue.get()

                # Special signal to close teh sender task
                if action == "CLOSE":
                    break
                
                try:
                    # Format and send the action
                    action_msg = {"command": "action", "action": action}
                    await self.websocket.send(json.dumps(action_msg))
                    print(f"Sent action: {action[:100]}...")

                    # Mark that we've sent an action and are waiting for a response
                    self.pending_action = True

                except Exception as e:
                    print(f"Error sending action: {e}")
                
                # Mark the task as done so that other coroutines waiting on .join() can proceed
                self.action_queue.task_done()

        except Exception as e:
            print(f"Action sender error: {e}")
            self.server_shutdown = True


    async def _ping_sender(self):
        """
        Background task to send periodic pings to the game server.

        This helps to keep the connection alive and detect if the server is still responsive.
        """
        try:
            while not self.server_shutdown:
                try:
                    # Send a ping message to the server
                    await self.websocket.send(json.dumps({"command": "ping"}))
                    await asyncio.sleep(25)
                except Exception as e:
                    print(f"Ping error: {e}")
                    break

        except Exception as e:
            print(f"Ping sender error: {e}")
            self.server_shutdown = True


    def _get_env_name_from_id(self, env_id):
        """Convert environment ID back to name for wrapper lookup."""
        # Create reverse mapping
        id_to_name = {v: k for k, v in NAME_TO_ID_DICT.items()}
        base_name = id_to_name.get(env_id)
        
        if base_name:
            # Check if any of our original env_id_names match this base name
            for env_name in self.env_id_names:
                if strip_env_variant(env_name) == base_name:
                    return env_name
        
        return base_name

    async def _process_matchmaking_message(self, message_str: str):
        """
        Handle a single message received from the matchmaking server.
        
        Depending on the 'command', this would update the queue status,
        complete the matchmaking, or handle errors. This is called by the matchmaking loop.
        """
        try:
            message = json.loads(message_str)
            command = message.get("command")
            
            if command == "queued":
                # Status: In queue
                avg_queue_time = message.get("avg_queue_time", 0)
                num_players = message.get("num_players_in_queue", 0)
                print(f"In queue. Average wait time: {avg_queue_time:.1f}s. Players in queue: {num_players}")
                
            elif command == "match_found":
                # Status: Match found - capture the game server details and environment ID
                self.game_url = message.get("game_url")
                self.env_id = message.get("env_id")  # This is the integer ID
                self.environment_id = message.get("environment_id")
                
                # Convert env_id back to name for wrapper application
                # The server returns env_id as string like "DontSayIt-v0", but we need to match it to our env_id_names
                server_env_name = message.get("env_id")  # This is actually the string name from server
                
                # Find the matching env_name from our original list
                self.matched_env_name = None
                for env_name in self.env_id_names:
                    if strip_env_variant(env_name) == strip_env_variant(server_env_name):
                        self.matched_env_name = env_name
                        break
                
                if not self.matched_env_name:
                    # Fallback to the server's env name
                    self.matched_env_name = server_env_name
                
                print(f"Match found! Environment: {self.matched_env_name} (Server ID: {server_env_name})")
                print(f"Connecting to game server: {self.game_url}")
                self.matchmaking_complete = True
                
            elif command == "error":
                # Status: Server-side error
                error_msg = message.get("message", "Unknown error")
                print(f"Matchmaking error: {error_msg}")
                
            elif command == "left":
                # Status: Client leaving the matchmaking queue
                print("Left matchmaking queue")
                
            else:
                print(f"Unknown matchmaking command: {command}")
                
        except json.JSONDecodeError:
            print(f"Invalid JSON received from matchmaking: {message_str}")

        except Exception as e:
            print(f"Error processing matchmaking message: {e}")

    async def connect_to_matchmaking(self):
        """
        Establish a WebSocket connection to the matchmaking server and queue for a game.

        This function:
        - Connects using the model's name and token (for identification/auth)
        - Sends a matchmaking 'queue' command with the desired environment(s)
        - Listens for queue updates and 'match_found'
        - Returns True if match was successful, False if timed out or errored
        """
        ssl_context = ssl.create_default_context()
        ssl_context.check_hostname = False
        ssl_context.verify_mode = ssl.CERT_NONE

        # Connect with model info for models
        query_params = {
            "model_name": self.model_name,
            "model_token": self.model_token,
        }
        query_string = urlencode(query_params)
        ws_uri = f"{MATCHMAKING_WS_URI}?{query_string}"
        
        print(f"Connecting to matchmaking server: {ws_uri}")
        
        try:
            # Create WebSocket connection
            self.matchmaking_websocket = await websockets.connect(
                ws_uri,
                # ssl=ssl_context,  # Uncomment for HTTPS
                ping_interval=20,
                ping_timeout=60
            )
            
            # Start background tasks for matchmaking
            asyncio.create_task(self._matchmaking_receiver())
            
            # Queue for a game
            queue_message = {
                "command": "queue",
                "environments": self.env_ids
            }
            await self.matchmaking_websocket.send(json.dumps(queue_message))
            print(f"Sent queue request for environments: {self.env_ids}")
            
            # Wait for match to be found or timeout
            start_time = asyncio.get_event_loop().time()
            while not self.matchmaking_complete:
                try:
                    # check for a new matchmaking message every 1 second
                    message = await asyncio.wait_for(
                        self.matchmaking_queue.get(),
                        timeout=1.0
                    )
                    await self._process_matchmaking_message(message)

                except asyncio.TimeoutError:
                    # Check if we should timeout the matchmaking
                    elapsed = asyncio.get_event_loop().time() - start_time
                    if elapsed > self.matchmaking_timeout:
                        print("Timeout waiting for match")
                        await self.matchmaking_websocket.close()
                        return False
                    continue
            
            # Match found - closing matchmaking websocket cleanly
            try:
                await self.matchmaking_websocket.close()
            except:
                pass
                
            return self.game_url is not None
            
        except Exception as e:
            print(f"Matchmaking connection error: {e}")
            return False

    async def connect_to_game_server(self):
        """
        Connect to the matched game server after matchmaking is complete. 

        Establishes a WebSocket connection and starts the background tasks for message handling.
        - _message_receiver: Receives messages from the game server
        - _action_sender: Sends actions to the game server
        - _ping_sender: Sends periodic pings to keep the connection alive

        Returns:
            bool: True if connected successfully, False otherwise
        """
        if not self.game_url:
            print("No game server IP available")
            return False
                
        ssl_context = ssl.create_default_context()
        ssl_context.check_hostname = False
        ssl_context.verify_mode = ssl.CERT_NONE
        
        # Initial delay to allow server initialization
        print("Waiting for game server to initialize...")
        await asyncio.sleep(2)

        ws_uri = f"wss://{self.game_url}/ws?token={self.model_token}"
        print(f"Connecting to game server: {ws_uri}")
        
        max_attempts = 15  # Increased from 10
        initial_grace_period = 12  # Don't show errors for first 8 seconds
        start_time = asyncio.get_event_loop().time()
        
        for attempt in range(1, max_attempts + 1):
            try:
                self.websocket = await websockets.connect(
                    ws_uri,
                    ssl=ssl_context,
                    ping_interval=30,
                    ping_timeout=90
                )
                
                asyncio.create_task(self._message_receiver())
                asyncio.create_task(self._action_sender())
                asyncio.create_task(self._ping_sender())
                
                elapsed_time = asyncio.get_event_loop().time() - start_time
                print(f"Connected to game server successfully after {elapsed_time:.1f}s")
                return True
                
            except Exception as e:
                elapsed_time = asyncio.get_event_loop().time() - start_time
                
                # Only show error messages after the grace period
                if elapsed_time > initial_grace_period:
                    print(f"Connection error (attempt {attempt}/{max_attempts}, {elapsed_time:.1f}s elapsed): {e}")
                else:
                    # During grace period, just show a waiting message occasionally
                    if attempt % 3 == 0:  # Every 3rd attempt during grace period
                        print(f"Waiting for server... ({elapsed_time:.1f}s elapsed)")
                
                if attempt < max_attempts:
                    # Adaptive delay: shorter delays initially, longer delays later
                    if elapsed_time < initial_grace_period:
                        delay = 1  # Quick retries during grace period
                    else:
                        delay = min(3, 1 + (attempt - 1) * 0.5)  # Gradually increase delay
                    
                    await asyncio.sleep(delay)
                else:
                    total_elapsed = asyncio.get_event_loop().time() - start_time
                    print(f"All connection attempts failed after {total_elapsed:.1f}s")
                    return False

    async def connect(self):
        """
        Connect to the matchmaking server and then to the game server.
        
        This function handles the entire connection process:
        - Connect to matchmaking server
        - Queue for a game
        - Connect to the game server once a match is found
        - Start background tasks for message handling and action sending
        
        Returns:
            bool: True if connected successfully, False otherwise
        """
        # First connect to matchmaking
        matchmaking_success = await self.connect_to_matchmaking()
        if not matchmaking_success:
            print("Failed to get a match")
            return False
            
        # Then connect to the game server
        return await self.connect_to_game_server()

    async def _process_message(self, message_str: str):
        """
        Handle a single message received from the game server websocket.

        Recognized commands include:
        - 'observation': Game state update (usually your turn)
        - 'game_over': End of the game with outcome and reward
        - 'timed_out': Someone failed to act in time
        - 'error': Server-side error
        - 'action_ack': Acknowledgement that action was received
        - 'ping': Server heartbeat request (respond with pong)
        - 'server_shutdown': Server has ended the session

        This is the central router for all game server-driven events.

        """
        try:
            message = json.loads(message_str)
            command = message.get("command")
            
            if command == "observation":
                # Received game state - this player's turn to act
                observation = message.get("observation")
                player_id = message.get("player_id")
                
                print(f"Received observation for player {player_id}")
                self.current_player_id = player_id
                self.current_observation = observation
                self.full_observations[player_id] = observation
                self.pending_action = False
                self.in_game = True

                self.step_info["player_id"] = player_id

                if player_id not in self.game_info:
                    self.game_info[player_id] = {"turn_count": 0, "invalid_move": False, "reason": ""}

                self.game_info[player_id]["turn_count"] += 1
                self.step_info["turn_count"] = self.game_info[player_id]["turn_count"]

                
            elif command == "game_over":
                # Game has completed - extract reason and any reward
                print("Game over received")
                self.game_over = True
                outcome = message.get("outcome", "unknown")
                reason = message.get("reason", "No reason provided")

                if self.current_player_id is not None:
                    self.rewards[self.current_player_id] = message.get("trueskill_change", 0)
                    self.game_info[self.current_player_id].update({
                        "reason": reason,
                        "outcome": outcome,
                        "invalid_move": self.game_info[self.current_player_id].get("invalid_move", False)
                    })

                self.step_info["game_end"] = True
                self.step_info["reason"] = reason
                self.step_info["outcome"] = outcome

                print(f"Game over: {outcome}, reason: {reason}")
                
            elif command == "timed_out":
                self.game_over = True
                timeout_msg = message.get("message", "Unknown timeout")

                if self.current_player_id is not None:
                    self.game_info[self.current_player_id].update({
                        "reason": "timeout",
                        "invalid_move": False
                    })

                self.step_info["timeout"] = True
                self.step_info["message"] = timeout_msg

                
            elif command == "error":
                error_msg = message.get("message", "Unknown error")
                print(f"Server error: {error_msg}")
                self.step_info["error"] = error_msg

                
            elif command == "action_ack":
                print("Action acknowledged by server")
                self.step_info["acknowledged"] = True
                
            elif command == "pong":
                pass
                
            elif command == "ping":
                try:
                    await self.websocket.send(json.dumps({"command": "pong"}))
                except Exception as e:
                    print(f"Error sending pong: {e}")
                    
            elif command == "server_shutdown":
                print("Server shutdown message received")
                self.server_shutdown = True
                
            else:
                print(f"Unknown command received: {command}")
                
        except json.JSONDecodeError:
            print(f"Invalid JSON received: {message_str}")

        except Exception as e:
            print(f"Error processing message: {e}")
            
    async def update_loop(self):
        """Main loop that processes messages."""
        game_over_time = None
        
        while not self.server_shutdown:
            try:
                timeout = 5.0 if self.game_over else None
                
                try:
                    # Wait for a message from the queue
                    # If game_over is set, wait for a message with a timeout
                    message = await asyncio.wait_for(self.message_queue.get(), timeout=timeout)
                    await self._process_message(message)
                    
                    # If this is the first game over, then start the timer
                    if self.game_over and game_over_time is None:
                        game_over_time = asyncio.get_event_loop().time()
                        print("Game over received, waiting for additional messages...")
                    
                except asyncio.TimeoutError:
                    # If we're in the post-game phase, then we track how long we've been waiting.
                    if self.game_over:
                        elapsed = asyncio.get_event_loop().time() - game_over_time
                        print(f"Timeout after {elapsed:.1f}s while waiting for additional messages after game over")
                        
                        if elapsed > self.game_over_timeout:
                            print(f"No more messages after {self.game_over_timeout}s of game over, exiting loop")
                            self.server_shutdown = True

                    else:
                        # Unexpected timeout, treating as a forced shutdown
                        print(f"Timeout while waiting for messages")
                        self.game_over = True
                        self.server_shutdown = True
                    
                # Check if we've waited long enough after game_over
                if self.game_over and game_over_time is not None:
                    elapsed = asyncio.get_event_loop().time() - game_over_time
                    if elapsed > self.game_over_timeout:
                        print(f"No more messages after {self.game_over_timeout}s of game over, exiting loop")
                        self.server_shutdown = True
                    
            except websockets.exceptions.ConnectionClosed:
                print("WebSocket connection closed by server")
                self.server_shutdown = True  # Set server_shutdown when connection is closed
                break
                
            except Exception as e:
                print(f"Error in update loop: {e}")
                await asyncio.sleep(0.1)
                
        print("Update loop exiting")

    async def async_get_observation(self) -> Tuple[Optional[int], List]:
        """
        Wait for and returns the current player's observation from the game server.

        If an observation is already available, it returns that immediately.
        Otherwise, it waits for an observation to be received, until either:
        - A valid observation is received
        - The server shuts down

        Returns:
            Tuple[player_id, observation], or (None, []) if timed out or invalid.
        """
        # If we already have an observation, return it
        if self.current_player_id is not None and self.current_observation:
            observation = self.current_observation
            player_id = self.current_player_id
            return player_id, observation

        if not self.server_shutdown:
            if self.update_task is None or self.update_task.done():
                self.update_task = asyncio.create_task(self.update_loop())
            
            try:
                # Wait until we get an observation or server shuts down
                start_time = asyncio.get_event_loop().time()

                while not self.server_shutdown:
                    if self.current_player_id is not None and self.current_observation:
                        return self.current_player_id, self.current_observation
                    await asyncio.sleep(0.1)
                        
            except Exception as e:
                print(f"Error waiting for observation: {e}")
                    
        self.observation_valid = False
        return None, []

    def get_observation(self) -> Tuple[Optional[int], List]:
        """
        Synchronous wrapper for async_get_observation, so non-async agents can call this.
        
        Handles asyncio event loop setup internally.
        Raises a RuntimeError if observation retrieval failed.
        """
        try:
            # get the current event loop (or create one)
            loop = asyncio.get_event_loop()
            if loop.is_closed():
                loop = asyncio.new_event_loop()
                asyncio.set_event_loop(loop)
                new_loop = True
            else:
                new_loop = False
        except RuntimeError:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            new_loop = True

        try:
            # run the async observation retrieval
            player_id, obs = loop.run_until_complete(self.async_get_observation())

            # Raise if invalid observation
            if getattr(self, "observation_valid", True) is False:
                raise RuntimeError("No valid observation — server shutdown or invalid state.")

            return player_id, obs
        
        finally:
            # Close the loop if we created a new one
            if new_loop:
                loop.close()

    async def async_step(self, action: str):
        """Take an action in the game."""
        if self.server_shutdown:
            return True, self.step_info

        self.step_info = {}

        if self.current_player_id is not None:
            self.step_info["player_id"] = self.current_player_id
            self.step_info["turn_count"] = self.game_info.get(self.current_player_id, {}).get("turn_count", 0) + 1

        await self.action_queue.put(action)
        await self.action_queue.join()
        self.current_observation = None

        while not self.server_shutdown and self.pending_action:
            await asyncio.sleep(0.1)

        return self.game_over, self.step_info

    def step(self, action: str):
        """
        Synchronous wrapper for async_step, so non-async agents can call this.

        Args:
            action: The action to be performed (e.g., "play x y", "bet 3")

        Returns:
            Tuple[bool, dict]: A tuple indicating if the game is over and any additional info
        """
        try:
            loop = asyncio.get_event_loop()
            if loop.is_closed():
                loop = asyncio.new_event_loop()
                asyncio.set_event_loop(loop)
                new_loop = True
            else:
                new_loop = False
        except RuntimeError:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            new_loop = True
            
        try:
            return loop.run_until_complete(self.async_step(action))
        finally:
            if new_loop:
                loop.close()

    async def async_reset(self, num_players=None, seed=None):
        """Connect to server and wait for game to start."""
        self.current_player_id = None
        self.current_observation = None
        self.game_over = False
        self.server_shutdown = False
        self.rewards = {}
        self.info = {}
        self.full_observations = {}
        self.in_game = False
        self.update_task = None
        self.matchmaking_complete = False
        
        # Connect to matchmaking server and game server if not already connected
        if not self.websocket:
            connected = await self.connect()
            if not connected:
                print("Failed to connect to server")
                await self.async_close()
                return []
                
        # Start the main message update loop
        self.update_task = asyncio.create_task(self.update_loop())
        
        try:
            # Wait until we either get an observation or the server shuts down
            start_time = asyncio.get_event_loop().time()
            while not self.server_shutdown and not self.in_game:
                await asyncio.sleep(0.1)
                
                if self.current_player_id is not None and self.current_observation:
                    self.in_game = True
                    return self.current_observation

        except Exception as e:
            print(f"Error waiting for game start: {e}")
                
        # Return current observation or empty list
        return self.current_observation if self.current_observation else []

    def reset(self, num_players=None, seed=None):
        """
        Synchronous wrapper for async_reset.

        Returns:
            The initial observation for the agent.
        """
        try:
            loop = asyncio.get_event_loop()
            if loop.is_closed():
                loop = asyncio.new_event_loop()
                asyncio.set_event_loop(loop)
                new_loop = True
            else:
                new_loop = False
        except RuntimeError:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            new_loop = True
            
        try:
            return loop.run_until_complete(self.async_reset(num_players))
        finally:
            if new_loop:
                loop.close()

    async def async_close(self):
        """
        Asynchronously close the environment and clean up resources.
        
        This function:
        - Signals the action sender to stop
        - Closes the game server websocket
        - Closes the matchmaking websocket if still open
        - Cancels the update loop task if running
        - Returns the rewards dictionary
        """
        # Set server_shutdown flag to ensure all loops terminate
        self.server_shutdown = True
        
        # Signal action sender to stop
        try:
            await self.action_queue.put("CLOSE")
        except:
            pass
            
        # Close game server websocket
        if self.websocket and not getattr(self.websocket, 'closed', True):
            try:
                await self.websocket.close()
            except:
                pass
            
        # Close matchmaking websocket if still open
        if self.matchmaking_websocket and not getattr(self.matchmaking_websocket, 'closed', True):
            try:
                await self.matchmaking_websocket.close()
            except:
                pass
                
        # Cancel update task if running
        if self.update_task and not self.update_task.done():
            try:
                self.update_task.cancel()
            except:
                pass
                
        return self.rewards, self.game_info

    def close(self):
        """
        Synchronous wrapper for async_close.

        This function handles the event loop setup and cleanup.

        Returns:
            The rewards dictionary from the last game.
        """
        try:
            loop = asyncio.get_event_loop()
            if loop.is_closed():
                loop = asyncio.new_event_loop()
                asyncio.set_event_loop(loop)
                new_loop = True
            else:
                new_loop = False

        except RuntimeError:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            new_loop = True
            
        try:
            return loop.run_until_complete(self.async_close())
        
        finally:
            if new_loop:
                loop.close()

def extract_agent_attributes(agent):
    """Extract relevant attributes from agent object for token generation."""
    if agent is None:
        return None
    
    return {
        "agent_class": agent.__class__.__name__,
        "agent_model": getattr(agent, "model_name", getattr(agent, "model_id", "")),
        "system_prompt": getattr(agent, "system_prompt", ""),
        "extra": getattr(agent, "kwargs", {}) or getattr(agent, "generation_config", {}) or {}
    }

def get_deterministic_model_token(
    email: str, model_name: str, agent_attributes: dict
) -> str:
    """Generate a deterministic UUID token based on email, model info, and agent config."""
    namespace = uuid.NAMESPACE_DNS
    agent_class = agent_attributes["agent_class"]
    agent_model = agent_attributes["agent_model"]
    system_prompt = agent_attributes["system_prompt"]
    extra_str = str(sorted(agent_attributes["extra"].items()))
    combined = f"{email}|{model_name}|{agent_class}|{agent_model}|{system_prompt}|{extra_str}"
    return str(uuid.uuid5(namespace, combined))

def register_model(model_name: str, description: str, email: str, agent_obj=None) -> str:
    """Register a model with the matchmaking server and get a token."""
    try:
        # Generate deterministic token if agent_obj is provided
        model_token = None
        agent_attributes = extract_agent_attributes(agent_obj)
        model_token = get_deterministic_model_token(email, model_name, agent_attributes)
        
        payload = {"model_name": model_name, "description": description, "email": email, "model_token": model_token}
        
        response = requests.post(
            f"{MATCHMAKING_HTTP_URI}/register_model",
            json=payload
        )
        
        # Handle different error cases with clear messages
        if response.status_code == 409:  # Conflict
            try:
                error_data = response.json()
                detail = error_data.get('detail', 'Model conflict')
                print(f"\n❌ Registration failed: {detail}")
                
                # Suggest specific solutions based on the error content
                if "email (existing:" in detail: print("💡 Solution: Use the exact same email as your previous registration")
                elif "different token" in detail: print("💡 Solution: Agent configuration changed - use a different model name or revert agent settings")
                else: print("💡 Suggestion: Try using a different model name")
                return None
            except: print(f"\n❌ Model already exists with different configuration."); return None
                
        elif response.status_code == 400:  # Bad Request
            try:
                error_data = response.json()
                detail = error_data.get('detail', 'Invalid request')
                print(f"\n❌ Registration failed: {detail}")
                # Handle specific 400 error cases
                if "Model token is required" in detail:
                    print("💡 Solution: Pass an agent object to make_online() to enable deterministic tokens:")
                    print("   Example:")
                    print("   agent = ta.agents.OpenRouterAgent(model_name='gpt-4o')")
                    print("   env = ta.make_online(..., agent_obj=agent)")
                elif "Invalid token format" in detail: print("💡 This is likely a bug - please report this issue")
                else: print("💡 Check your request parameters and try again")
                return None
            except: print(f"\n❌ Invalid request: {response.text}"); return None
                
        elif response.status_code != 200: print(f"\n❌ Server error ({response.status_code}): {response.text}"); return None
        
        # Success case
        response.raise_for_status()
        data = response.json()
        return data.get("model_token")
        
    except requests.exceptions.RequestException as e:
        print(f"\n❌ Network error registering model: {e}")
        return None

    except Exception as e:
        print(f"\n❌ Unexpected error registering model: {e}")
        return None


def make_online(
    env_id: Union[str, List[str]],
    model_name: str,
    model_token: Optional[str] = None,
    model_description: Optional[str] = None,
    email: Optional[str] = None,
    agent: Optional[object] = None,
) -> Union[OnlineEnvWrapper, DynamicWrapperProxy]:
    """Create and return an online environment with appropriate wrappers."""

    # Ensure env_ids is a list
    env_ids = [env_id] if isinstance(env_id, str) else env_id

    # Convert to internal numeric env IDs
    env_ids_int = []
    for full_id in env_ids:
        base_id = strip_env_variant(full_id)
        if base_id not in NAME_TO_ID_DICT:
            raise ValueError(f"Environment {full_id} not recognized (base: {base_id})")
        env_ids_int.append(NAME_TO_ID_DICT[base_id])

    # Handle model registration
    if not model_token:
        if not email or not agent:
            raise ValueError("Provide email and agent if model_token is not given.")
        model_token = register_model(model_name, model_description, email, agent)
        if not model_token:
            raise ValueError("Model registration failed.")
        print(f"✅ Registered '{model_name}' with {'deterministic' if agent else 'random'} token: {model_token}")
    else:
        print(f"✅ Using provided token for '{model_name}': {model_token}")

    # Create base wrapper
    base_env = OnlineEnvWrapper(env_ids_int, env_ids, model_name, model_token)

    # Collect default wrappers
    env_id_to_wrappers = {}
    for name in env_ids:
        spec = ENV_REGISTRY.get(name)
        wrappers = spec.default_wrappers if spec and spec.default_wrappers else []
        env_id_to_wrappers[name] = wrappers
        if not spec:
            print(f"[make_online] Warning: '{name}' not found in ENV_REGISTRY")

    # Pretty log: Table format
    print(f"{'Environment':<30} | Wrappers")
    print("-" * 70)
    for name in sorted(env_id_to_wrappers):
        wrappers = env_id_to_wrappers[name]
        wrapper_names = ", ".join(w.__name__ for w in wrappers) if wrappers else "None"
        print(f"{name:<30} | {wrapper_names}")
    print()

    # Apply immediately if single environment
    if len(env_ids) == 1 and env_ids[0] in ENV_REGISTRY:
        wrappers = env_id_to_wrappers[env_ids[0]]
        if wrappers:
            print(f"[make_online] Applying wrappers for '{env_ids[0]}':")
            for wrapper in wrappers:
                print(f"  - {wrapper.__name__}")
                base_env = wrapper(base_env)
        return base_env

    # Multi-env setup → return dynamic proxy
    return DynamicWrapperProxy(base_env, env_id_to_wrappers)

#### Mind Games Challenge (mgc) specific code ####

MGC_NAME_TO_ID_DICT = { # TODO change to the correct mapping
    "ConnectFour-v0": 1,
    "DontSayIt-v0": 3,
    "Nim-v0": 50,
    "Snake-v0": 69,
}

## create a custom make_online for the competition of mindgameschallenge
def make_mgc_online(
    env_id: Union[str, List[str]],
    model_name: str,
    model_token: Optional[str] = None,
    model_description: Optional[str] = None,
    team_hash: Optional[str] = None,
    agent: Optional[object] = None,
) -> OnlineEnvWrapper:
    """ Create and return an online environment for Mind Games Challenge (mgc) for the game environments of 1,2,3,4 """

    # Ensure env_ids is a list
    env_ids = [env_id] if isinstance(env_id, str) else env_id

    # Convert to internal numeric env IDs
    env_ids_int = []
    for full_id in env_ids:
        base_id = strip_env_variant(full_id)
        if base_id not in MGC_NAME_TO_ID_DICT:
            raise ValueError(f"Environment {full_id} not recognized (base: {base_id} for MindGamesChallenge)")
        env_ids_int.append(MGC_NAME_TO_ID_DICT[base_id])

    # Handle model registration
    if not model_token:
        if not team_hash or not agent:
            raise ValueError("Provide email and agent if model_token is not given.")
        model_token = register_model(model_name, model_description, team_hash, agent)

        if not model_token:
            raise ValueError("Model registration failed.")
        print(f"✅ Registered '{model_name}' with {'deterministic' if agent else 'random'} token: {model_token}")
    else:
        print(f"✅ Using provided token for '{model_name}': {model_token}")

    # Create base wrapper
    base_env = OnlineEnvWrapper(env_ids_int, env_ids, model_name, model_token)

    # Collect default wrappers
    env_id_to_wrappers = {}
    for name in env_ids:
        spec = ENV_REGISTRY.get(name)
        wrappers = spec.default_wrappers if spec and spec.default_wrappers else []
        env_id_to_wrappers[name] = wrappers
        if not spec:
            print(f"[make_online] Warning: '{name}' not found in ENV_REGISTRY")

    # Pretty log: Table format
    print(f"{'Environment':<30} | Wrappers")
    print("-" * 70)
    for name in sorted(env_id_to_wrappers):
        wrappers = env_id_to_wrappers[name]
        wrapper_names = ", ".join(w.__name__ for w in wrappers) if wrappers else "None"
        print(f"{name:<30} | {wrapper_names}")
    print()

    # Apply immediately if single environment
    if len(env_ids) == 1 and env_ids[0] in ENV_REGISTRY:
        wrappers = env_id_to_wrappers[env_ids[0]]
        if wrappers:
            print(f"[make_online] Applying wrappers for '{env_ids[0]}':")
            for wrapper in wrappers:
                print(f"  - {wrapper.__name__}")
                base_env = wrapper(base_env)
        return base_env

    # Multi-env setup → return dynamic proxy
    return DynamicWrapperProxy(base_env, env_id_to_wrappers)


import re
from typing import Optional, Dict, Tuple, Any

import textarena as ta
from textarena.envs.ThreePlayerTicTacToe.renderer import create_board_str

class ThreePlayerTicTacToeEnv(ta.Env):
    def __init__(self):
        super().__init__()
        self.board_size = 5
        self.cell_mapping = {i * self.board_size + j: (i, j) for i in range(self.board_size) for j in range(self.board_size)}
        self.symbols = {0: 'A', 1: 'B', 2: 'C'}

    def get_board_str(self): return create_board_str(game_state=self.state.game_state)
    def reset(self, num_players: int, seed: Optional[int] = None):
        assert num_players==3, f"ThreePlayerTicTacToe only works with exactly three players. {num_players} players were provided."
        self.state = ta.FFAMultiPlayerState(num_players=num_players, seed=seed)
        self.state.reset(game_state={"board": [['' for _ in range(self.board_size)] for _ in range(self.board_size)]}, player_prompt_function=self._prompt)
        self._observer_current_state()

    def _prompt(self, player_id: int, game_state: Dict[str, Any]) -> str:
        return (
            f"You are Player {player_id} in Three-Player Tic Tac Toe.\nYour symbol is '{self.symbols[player_id]}'.\n"
            "You take turns placing your symbol on a 5x5 board to form a line of four.\nLines can be horizontal, vertical, or diagonal.\n"
            "Submit your move using the format '[4]' to mark cell 4."
        )
    
    def _render_board(self) -> str:
        cell_width = max(len(str(self.board_size * self.board_size - 1)), 2)  # ensures symbols like 'A', 'B', 'C' are well-centered
        def cell_str(r: int, c: int) -> str: return self.state.game_state["board"][r][c] if self.state.game_state["board"][r][c] != '' else str(r * self.board_size + c)
        def build_hline() -> str: return "+" + "+".join("-" * (cell_width + 2) for _ in range(self.board_size)) + "+"
        lines = [build_hline()]
        for r in range(self.board_size):
            row_cells = [f" {cell_str(r, c):^{cell_width}} " for c in range(self.board_size)]
            row_line = "|" + "|".join(row_cells) + "|"
            lines.append(row_line)
            lines.append(build_hline())
        return "\n".join(lines)

    def _observer_current_state(self) -> None:
        available_moves = []
        for i in range(self.board_size * self.board_size):
            r, c = self.cell_mapping[i]
            if self.state.game_state["board"][r][c] == '':
                available_moves.append(f"[{i}]")
        self.state.add_observation(message=f"Current Board:\n\n{self._render_board()}\n\nAvailable Moves: " + ", ".join(available_moves), observation_type=ta.ObservationType.GAME_BOARD)

    def step(self, action: str) -> Tuple[bool, ta.Info]:
        player_id = self.state.current_player_id
        self.state.add_observation(from_id=player_id, message=action, observation_type=ta.ObservationType.PLAYER_ACTION)
        match = re.search(r"\[(\d+)\]", action)
        if not match:
            if self.state.set_invalid_move(reason=f"Invalid move format. Use '[cell]' where cell is 0-{self.board_size ** 2 - 1}."):
                self.state.set_winners(player_ids=[pid for pid in range(3) if pid!=player_id], reason=f"Player {player_id} made an invalid move")
        else:
            cell = int(match.group(1))
            if cell not in self.cell_mapping:
                if self.state.set_invalid_move(reason=f"Invalid cell number: {cell}. Must be between 0 and {self.board_size ** 2 - 1}."):
                    self.state.set_winners(player_ids=[pid for pid in range(3) if pid!=player_id], reason=f"Player {player_id} made an invalid move")
            else:
                row, col = self.cell_mapping[cell]
                if self.state.game_state["board"][row][col] == '':
                    self.state.game_state["board"][row][col] = self.symbols[player_id]
                    self.state.add_observation(message=f"Player {player_id} places their symbol ({self.symbols[player_id]}) in field ({row}, {col}).", observation_type=ta.ObservationType.GAME_ACTION_DESCRIPTION)
                    if self._check_winner(self.symbols[player_id]): 
                        self.state.set_winners(player_ids=[player_id], reason=f"Player {player_id} ({self.symbols[player_id]}) wins!")
                    elif all(cell != '' for row in self.state.game_state["board"] for cell in row): self.state.set_draw(reason="The game is a draw!")
                else:
                    if self.state.set_invalid_move(reason=f"Cell {cell} is already occupied."): self.state.set_winners(player_ids=[pid for pid in range(3) if pid!=player_id], reason=f"Player {player_id} made an invalid move")
        self._observer_current_state()
        return self.state.step()

    def _check_winner(self, symbol: str) -> bool:
        win_length = 4
        # Horizontal & Vertical
        for r in range(self.board_size):
            for c in range(self.board_size - win_length + 1):
                if all(self.state.game_state["board"][r][c + i] == symbol for i in range(win_length)): return True
        for c in range(self.board_size):
            for r in range(self.board_size - win_length + 1):
                if all(self.state.game_state["board"][r + i][c] == symbol for i in range(win_length)): return True
        # Diagonal \ and /
        for r in range(self.board_size - win_length + 1):
            for c in range(self.board_size - win_length + 1):
                if all(self.state.game_state["board"][r + i][c + i] == symbol for i in range(win_length)): return True
                if all(self.state.game_state["board"][r + i][c + win_length - 1 - i] == symbol for i in range(win_length)): return True
        return False


def create_board_str(game_state: dict) -> str:
    """
    Create a board representing Hanabi.

    Args:
        game_state (dict): A dictionary containing the game state.

    Returns:
        str: A board representing Hanabi.
    """
    fireworks = game_state.get("fireworks")
    deck_size = len(game_state.get("deck", []))
    discard_pile = game_state.get('discard_pile', [])
    info_tokens = game_state.get('info_tokens', 8)
    fuse_tokens = game_state.get('fuse_tokens', 4)


    lines = []

    # Phase display
    lines.append(f"╭── Hanabi ──────────────────────────────────────────────────╮")
    lines.append(f"│   Deck size: {deck_size:<2}   │   Info tokens: {info_tokens:<1}   │  Fuse tokens: {fuse_tokens:<1}   │")
    lines.append(f"╰────────────────────────────────────────────────────────────╯")

    # Scoreboard
    lines.append("┌─ 🎆 Fireworks ─────────────────────────────────────────────┐")
    for key in fireworks.keys():
        firework = "D" + "=" * int(fireworks[key]) + ">" + " " *(10-int(fireworks[key]))
        lines.append(f"│ {key.value:<10}: {fireworks[key]:<2} │ {firework}                              │")
    lines.append("└────────────────────────────────────────────────────────────┘")

    # Discard pile
    lines.append("┌─ 🚮 Discard pile ──────────────────────────────────────────┐")
    for card in discard_pile:
        lines.append(f"│ {str(card):<59}│")
    lines.append("└────────────────────────────────────────────────────────────┘")

    return "\n".join(lines)